# Copyright 2025 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Disentangled RNN and plotting functions."""
import dataclasses
from typing import Optional, Callable, Any, Sequence

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np


def information_bottleneck(
    inputs: jnp.ndarray,
    sigmas: jnp.ndarray,
    multipliers: Optional[jnp.ndarray] = None,
    noiseless_mode: bool = False,
) -> tuple[jnp.ndarray, jnp.ndarray]:
  r"""Output from an information bottleneck given a vector of means and std devs.

  Bottleneck outputs are sampled independently from Gaussian distributions with
  the given means and variances. Bottleneck costs are computed as the KL
  divergence between this sampling distribution and the unit Gaussian.

  KL(p, q) = H(p, q) - H(p) = -\int p(x)log(q(x))dx - -\int p(x)log(p(x))dx
           = 0.5 * [log(|s2|/|s1|) - 1 + tr(s1/s2) + (m1-m2)^2/s2]
           = 0.5 * [-log(|s1|) - 1 + tr(s1) + m1^2] (if m2 = 0, s2 = 1)

  Args:
    inputs: The inputs to the bottleneck. Shape is (batch_size, bottleneck_dims)
    sigmas: The standard deviations of the sampling distribution (diagonal of
      the sqrt of the covariance matrix). Shape is (bottleneck_dims).
    multipliers: The multipliers to apply to the inputs. Shape is
      (batch_size, bottleneck_dims)
    noiseless_mode: If True, no noise is added and no penalty is computed.

  Returns:
    bottleneck_output: The noisy output of the bottleneck. Shape is the same as
      mus.
    bottleneck_penalty: The KL cost of the bottleneck sample. Shape is
      (batch_size,).
  """
  # Apply multipliers if any are provided
  # Do this regardless of whether we are in noiseless mode, so the average
  # output will be the same
  if multipliers is None:
    mus = inputs
  else:
    mus = inputs * multipliers

  if noiseless_mode:
    # In noiseless mode, do add noise and do not include penalty
    bottleneck_output = mus
    bottleneck_penalty = jnp.zeros(shape=(mus.shape[0],))
  else:
    # In noisey mode, add noise based on sigmas and compute KL penalty
    # Shape is (batch_size, bottleneck_dims)
    bottleneck_output = mus + sigmas * jax.random.normal(
        hk.next_rng_key(), jnp.shape(mus)
    )

    # KL divergence KL( N(mu, var) || N(0, 1))
    variances = jnp.square(sigmas)
    # Shape is (batch_size, bottleneck_dims)
    elementwise_kl = jnp.square(mus) + variances - 1.0 - jnp.log(variances)
    # Shape is (batch_size,)
    bottleneck_penalty = 0.5 * jnp.sum(
        elementwise_kl, axis=tuple(range(1, elementwise_kl.ndim))
    )

  return bottleneck_output, bottleneck_penalty


def reparameterize_sigma(
    hk_param: jnp.ndarray, min_sigma: float = 1e-5
) -> jnp.ndarray:
  """Reparamaterizes bottleneck sigma for easy fitting.

  Args:
    hk_param: The haiku parameter corresponding to a bottleneck sigma. Range
      from -inf to +inf
    min_sigma: The minimum value of the standard deviation.
  Returns:
    sigma: The bottleneck standard deviation. Range from min_sigma to inf.
  """
  return jnp.abs(hk_param) + min_sigma


@dataclasses.dataclass
class DisRnnConfig:
  """Specifies an architecture and configuration for a Disentangled RNN.

  Attributes:
    obs_size: Number of dimensions in the observation vector
    output_size: Number of dimensions the disRNN will output
      (logits or predicted targets)
    latent_size: Number of recurrent variables
    update_net_n_units_per_layer: Number of units in each layer of the update
      networks
    update_net_n_layers: Number of layers in the update networks
    choice_net_n_units_per_layer: Number of units in each layer of the choice
      network
    choice_net_n_layers: Number of layers in the choice network
    noiseless_mode: Allows turning off the bottlenecks e.g. for evaluation
    latent_penalty: Multiplier for KL cost on the latent bottlenecks
    choice_net_latent_penalty: Multiplier for bottleneck cost on latent inputs
      to the choice network
    update_net_obs_penalty: Multiplier for bottleneck cost on observation
      inputs to the update network
    update_net_latent_penalty: Multiplier for latent inputs to the update
      networks
    l2_scale: Multiplier for L2 penalty on hidden layer weights in both update
      and choice networks
    activation: String defining an activation function. Must be in jax.nn.
    max_latent_value: Cap on the possible absolute value of a latent. Used to
      prevent runaway latents resulting in NaNs
    x_names: Names of the observation vector elements. Must have length obs_size
    y_names: Names of the target vector elements. Must have length target_size
  """

  obs_size: int = 2
  output_size: int = 2
  latent_size: int = 10

  update_net_n_units_per_layer: int = 10
  update_net_n_layers: int = 2
  choice_net_n_units_per_layer: int = 2
  choice_net_n_layers: int = 2
  activation: str = 'leaky_relu'

  noiseless_mode: bool = False

  latent_penalty: float = 1.0
  update_net_obs_penalty: float = 0.0
  update_net_latent_penalty: float = 0.0
  choice_net_latent_penalty: float = 0.0

  l2_scale: float = 0.01

  max_latent_value: float = 2.

  x_names: Optional[list[str]] = None
  y_names: Optional[list[str]] = None

  def __post_init__(self):
    """Checks that the configuration is valid."""

    expected_len_xnames = self.obs_size
    if self.x_names is None:
      self.x_names = [f'Observation {i}' for i in range(expected_len_xnames)]
    if len(self.x_names) != expected_len_xnames:
      raise ValueError(
          f'Based on obs_size {self.obs_size}, expected x_names to have '
          f'length {self.obs_size} but got {self.x_names}'
      )

    # Check activation is in jax.nn
    try:
      getattr(jax.nn, self.activation)
    except AttributeError as e:
      raise ValueError(
          f'Activation {self.activation} not found in jax.nn. Provided value '
          f'was {self.activation}'
      ) from e


class ResMLP(hk.Module):
  """MLP modified to apply serial updates to a residual stream.

  Attributes:
    input_size: Dimension of the input vector
    output_size: Dimensions of the output vector
    n_layers: Number of layers
    n_units_per_layer: Dimension of the stream and of each layer
    activation_fn: Activation function
    w_initializer: Initializer for the weights
    b_initializer: Initializer for the biases
    name: Optional name, which affects the names of the haiku parameters
  """

  def __init__(self,
               input_size: int,
               output_size: int,
               n_layers: int = 5,
               n_units_per_layer: int = 5,
               activation_fn: Callable[[Any], Any] = jax.nn.relu,
               name=None):
    super().__init__(name=name)

    self.n_layers = n_layers
    self.n_units_per_layer = n_units_per_layer
    self.activation_fn = activation_fn

    # Input layer will be a linear projection from input size to stream size
    # To keep activation magnitudes similar, we initialise weights in the range
    # 1/sqrt(input_size)
    scale = 1 / jnp.sqrt(jnp.float32(input_size))
    self._input_weights = hk.get_parameter(
        'input_weights',
        (input_size, self.n_units_per_layer),
        init=hk.initializers.RandomNormal(stddev=scale),
    )
    self._input_biases = hk.get_parameter(
        'input_biases',
        (self.n_units_per_layer,),
        init=hk.initializers.Constant(0.0),
    )

    # Output layer will be a linear projection from stream size to output size,
    # To keep activation magnitudes similar, we scale initial weights by
    # 1 / sqrt(n_units_per_layer)
    scale = 1 / jnp.sqrt(jnp.float32(self.n_units_per_layer))
    self._output_weights = hk.get_parameter(
        'output_weights',
        (n_units_per_layer, output_size),
        init=hk.initializers.RandomNormal(stddev=scale),
    )
    self._output_biases = hk.get_parameter(
        'output_biases', (output_size,), init=hk.initializers.Constant(0.0)
    )

    # Hidden layers will each be a single fully connected layer.
    # Each layer will increase the variance of the stream, so we scale initial
    # weights both by n_units_per_layer and n_layers.
    hidden_w_init_scale = 1 / jnp.sqrt(
        jnp.float32(n_units_per_layer * n_layers)
    )
    hidden_w_init = hk.initializers.RandomNormal(stddev=hidden_w_init_scale)
    self._hidden_layer_weights = []
    self._hidden_layer_biases = []
    for hidden_layer_i in range(self.n_layers):
      self._hidden_layer_weights.append(
          hk.get_parameter(
              f'layer_{hidden_layer_i}_weights',
              (self.n_units_per_layer, self.n_units_per_layer),
              init=hidden_w_init,
          )
      )
      self._hidden_layer_biases.append(
          hk.get_parameter(
              f'layer_{hidden_layer_i}_biases',
              (self.n_units_per_layer,),
              init=hk.initializers.Constant(0.0),
          )
      )

    # Compute sum of squares of all hidden layer weights. This will be passed on
    # and can be used to compute an L2 (ridge) penalty.
    self.l2 = (
        jnp.sum(jnp.square(jnp.array(self._hidden_layer_weights)))
    )

  def __call__(self, inputs):

    # Linear projection of inputs to the size of the residual stream
    stream = jnp.dot(inputs, self._input_weights) + self._input_biases

    # Each iteration adds a layer.
    # Each hidden layer additively modifies the residual stream.
    for hidden_layer_i in range(self.n_layers):
      # (batch_size, stream_size)
      layer_activations = (
          jnp.dot(stream, self._hidden_layer_weights[hidden_layer_i])
          + self._hidden_layer_biases[hidden_layer_i]
      )
      layer_output = self.activation_fn(layer_activations)
      stream += layer_output
    # Linear projection to the appropriate output size
    output = jnp.dot(stream, self._output_weights) + self._output_biases

    return output, self.l2


def get_initial_bottleneck_params(
    shape: Sequence[int], name: str,
) -> tuple[jnp.ndarray, jnp.ndarray]:
  """Defines a bottleneck with a sigma and a multiplier."""
  # At init the bottlenecks should all be open: sigmas small and multipliers 1
  sigma_params = hk.get_parameter(
      name + '_sigma_params',
      shape,
      init=hk.initializers.RandomUniform(minval=0.0, maxval=0.05),
  )
  sigmas = reparameterize_sigma(sigma_params)
  multipliers = hk.get_parameter(
      name + '_multipliers',
      shape,
      init=hk.initializers.Constant(constant=1),
  )
  return sigmas, multipliers


class HkDisentangledRNN(hk.RNNCore):
  """Disentangled RNN."""

  def __init__(
      self,
      config: DisRnnConfig,
  ):
    super().__init__()
    self._l2_scale = config.l2_scale
    self._noiseless_mode = config.noiseless_mode

    self._obs_size = config.obs_size
    self._output_size = config.output_size
    self._latent_size = config.latent_size

    self._update_net_n_units_per_layer = config.update_net_n_units_per_layer
    self._update_net_n_layers = config.update_net_n_layers
    self._choice_net_n_units_per_layer = config.choice_net_n_units_per_layer
    self._choice_net_n_layers = config.choice_net_n_layers

    self._latent_penalty = config.latent_penalty
    self._update_net_obs_penalty = config.update_net_obs_penalty
    self._update_net_latent_penalty = config.update_net_latent_penalty
    self._choice_net_latent_penalty = config.choice_net_latent_penalty
    self._activation = getattr(jax.nn, config.activation)
    self._max_latent_value = config.max_latent_value

    # Get Haiku parameters. IMPORTANT: if you are subclassing HkDisentangledRNN,
    # you must override _get_haiku_parameters to add any new parameters that you
    # need. This way of doing things is necessary for Haiku to work correctly.
    self._get_haiku_parameters()

  def _get_haiku_parameters(self):
    """Initializes parameters for the bottlenecks."""
    self._build_update_bottlenecks()
    self._build_latent_bottlenecks()
    self._build_choice_bottlenecks()
    self._latent_inits = hk.get_parameter(
        'latent_inits',
        (self._latent_size,),
        init=hk.initializers.RandomUniform(minval=-0.1, maxval=0.1),
    )

  def _build_update_bottlenecks(self):
    """Initializes parameters for the update network bottlenecks."""
    # There is one Update Network per latent. Each one gets input from all
    # latents and all observations. These inputs pass through separate
    # bottlenecks.
    # Bottlenecks for observations feeding into update nets
    self._update_net_obs_sigmas, self._update_net_obs_multipliers = (
        get_initial_bottleneck_params(
            shape=(self._obs_size, self._latent_size),
            name='update_net_obs',
        )
    )
    # Bottlenecks for previous latents feeding into update nets
    self._update_net_latent_sigmas, self._update_net_latent_multipliers = (
        get_initial_bottleneck_params(
            shape=(self._latent_size, self._latent_size),
            name='update_net_latent',
        )
    )

  def _build_latent_bottlenecks(self):
    """Initializes parameters for the latent bottlenecks."""
    # Latents will also go through a bottleneck after being updated. These
    # bottlenecks do not need multipliers, the network output can rescale them
    self._latent_sigmas, _ = (
        get_initial_bottleneck_params(
            shape=(self._latent_size,),
            name='latent',
        )
    )

  def _build_choice_bottlenecks(self):
    """Initializes parameters for the choice network bottlenecks."""
    # Choice network gets inputs from the latents, and has a bottleneck on each
    self._choice_net_sigmas, self._choice_net_multipliers = (
        get_initial_bottleneck_params(
            shape=(self._latent_size,),
            name='choice_net',
        )
    )

  def initial_state(self, batch_size: Optional[int]) -> Any:
    # (batch_size, latent_size)
    latents = jnp.ones([batch_size, self._latent_size]) * self._latent_inits
    return latents

  def update_latents(
      self, update_net_inputs: jnp.ndarray, prev_latent_values: jnp.ndarray
  ) -> tuple[jnp.ndarray, jnp.ndarray]:
    """Updates the latents using the update rules.

    Each latent is updated by a separate Update Network, which takes as input
    all previous latents and all additional update rule inputs, and outputs an
    update and a weight. New value of the latent will be the weighted average of
    the previous latent and the update.

    Args:
      update_net_inputs: Additional inputs for the update rules.
      prev_latent_values: The latents from the previous time step.
    Returns:
      new_latent_values: The updated latents.
      penalty_increment: A penalty associated with the update.
    """
    # penalty_increment: (batch_size,)
    batch_size = prev_latent_values.shape[0]
    penalty_increment = jnp.zeros(shape=(batch_size,))

    # Loop over latents. Update each using its own network.
    new_latent_values = jnp.zeros(
        shape=(prev_latent_values.shape[0], self._latent_size)
    )
    for net_i in jnp.arange(self._latent_size):
      update_net_output, update_net_l2 = ResMLP(
          input_size=update_net_inputs.shape[1],
          output_size=2,
          n_units_per_layer=self._update_net_n_units_per_layer,
          n_layers=self._update_net_n_layers,
          activation_fn=self._activation,
          name='update_net',
      )(update_net_inputs[:, :, net_i])
      # Add L2 to the penalty based on weights of the network
      penalty_increment += self._l2_scale * update_net_l2
      # Update the latent multiplicatively, using a weight and a new target
      # derived from the MLP output.
      # Weight needs to be in [0, 1]
      new_latent_weight = jax.nn.sigmoid(update_net_output[:, 0])
      # Target needs to be in [-max_latent_value, max_latent_value]
      new_latent_target = update_net_output[:, 1]
      new_latent_target = self._max_latent_value * jax.nn.tanh(
          new_latent_target / self._max_latent_value
      )
      # New latent value is weighted average of previous value and new target.
      prev_latent_value = prev_latent_values[:, net_i]
      new_latent_value = (
          1 - new_latent_weight
      ) * prev_latent_value + new_latent_weight * new_latent_target
      new_latent_values = new_latent_values.at[:, net_i].set(new_latent_value)

    # Put latent values through an information bottleneck. If we are in
    # noiseless mode, we can skip this.
    new_latent_values, latent_kl_cost = information_bottleneck(
        inputs=new_latent_values,
        sigmas=self._latent_sigmas,
        noiseless_mode=self._noiseless_mode,
    )
    penalty_increment += self._latent_penalty * latent_kl_cost

    return new_latent_values, penalty_increment

  def predict_targets(
      self,
      choice_net_inputs: jnp.ndarray,
  ) -> tuple[jnp.ndarray, jnp.ndarray]:
    """Predicts the targets using the choice network."""
    batch_size = choice_net_inputs.shape[0]
    penalty_increment = jnp.zeros(shape=(batch_size,))

    predicted_targets, choice_net_l2 = ResMLP(
        input_size=choice_net_inputs.shape[1],
        output_size=self._output_size,
        n_units_per_layer=self._choice_net_n_units_per_layer,
        n_layers=self._choice_net_n_layers,
        activation_fn=self._activation,
        name='choice_net'
        )(choice_net_inputs)
    penalty_increment += self._l2_scale * choice_net_l2

    return predicted_targets, penalty_increment

  def __call__(self, observations: jnp.ndarray, prev_latents: jnp.ndarray):
    # Initial penalty values. Shape is (batch_size,)
    batch_size = prev_latents.shape[0]
    penalty = jnp.zeros(shape=(batch_size,))

    # Prepare observations for update nets
    # Expand to have a separate copy per update network:
    # (batch_size, obs_size, latent_size)
    obs_for_update_net = jnp.tile(
        jnp.expand_dims(observations, 2), (1, 1, self._latent_size)
    )
    obs_for_update_net, obs_update_kl = information_bottleneck(
        inputs=obs_for_update_net,
        sigmas=self._update_net_obs_sigmas,
        multipliers=self._update_net_obs_multipliers,
        noiseless_mode=self._noiseless_mode,
    )
    penalty += self._update_net_obs_penalty * obs_update_kl

    # Prepare previous latents for update nets
    # Expand to have a separate copy per update network:
    # (batch_size, latent_size, latent_size)
    prev_latents_for_update_net = jnp.tile(
        jnp.expand_dims(prev_latents, 2), (1, 1, self._latent_size)
    )
    prev_latents_for_update_net, latent_update_kl = information_bottleneck(
        inputs=prev_latents_for_update_net,
        sigmas=self._update_net_latent_sigmas,
        multipliers=self._update_net_latent_multipliers,
        noiseless_mode=self._noiseless_mode,
    )
    penalty += self._update_net_latent_penalty * latent_update_kl

    # Concatenate processed inputs for the update networks
    # Shape: (batch_size, obs_size + latent_size, latent_size)
    update_net_inputs = jnp.concatenate(
        (obs_for_update_net, prev_latents_for_update_net), axis=1
    )

    # Update latents using update networks
    new_latents, penalty_increment = self.update_latents(
        update_net_inputs, prev_latents
    )
    penalty += penalty_increment

    # Set up choice net inputs from new_latents.
    choice_net_inputs, choice_net_input_kl = information_bottleneck(
        inputs=new_latents,
        sigmas=self._choice_net_sigmas,
        multipliers=self._choice_net_multipliers,
        noiseless_mode=self._noiseless_mode,
    )
    penalty += self._choice_net_latent_penalty * choice_net_input_kl

    predicted_targets, penalty_increment = self.predict_targets(
        choice_net_inputs
    )
    penalty += penalty_increment

    # Output has shape (batch_size, output_size + 1).
    # The first output_size elements are the predicted targets, and the last
    # element is the penalty. We preassign instead of using concatenate to avoid
    # errors caused by silent broadcasting.
    output_shape = (batch_size, self._output_size + 1)
    output = jnp.zeros(output_shape)
    output = output.at[:, :-1].set(predicted_targets)
    output = output.at[:, -1].set(penalty)

    return output, new_latents


def log_bottlenecks(params,
                    open_thresh: float = 0.1,
                    partially_open_thresh: float = 0.25,
                    closed_thresh: float = 0.9) -> dict[str, int]:
  """Computes info about bottlenecks for the base DisRNN."""

  params_disrnn = params['hk_disentangled_rnn']

  latent_sigmas = np.array(
      reparameterize_sigma(params_disrnn['latent_sigma_params'])
  )
  update_obs_sigmas = np.array(
      reparameterize_sigma(
          np.transpose(params_disrnn['update_net_obs_sigma_params'])
      )
  )
  update_latent_sigmas = np.array(
      reparameterize_sigma(
          np.transpose(params_disrnn['update_net_latent_sigma_params'])
      )
  )
  choice_sigmas = np.array(
      reparameterize_sigma(
          np.transpose(params_disrnn['choice_net_sigma_params'])
      )
  )

  latent_bottlenecks_open = np.sum(latent_sigmas < open_thresh)
  choice_bottlenecks_open = np.sum(choice_sigmas < open_thresh)
  update_obs_bottlenecks_open = np.sum(update_obs_sigmas < open_thresh)
  update_latent_bottlenecks_open = np.sum(update_latent_sigmas < open_thresh)
  update_bottlenecks_open = (
      update_obs_bottlenecks_open + update_latent_bottlenecks_open
  )

  latent_bottlenecks_partial = np.sum(latent_sigmas < partially_open_thresh)
  choice_bottlenecks_partial = np.sum(choice_sigmas < partially_open_thresh)
  update_obs_bottlenecks_partial = np.sum(
      update_obs_sigmas < partially_open_thresh
  )
  update_latent_bottlenecks_partial = np.sum(
      update_latent_sigmas < partially_open_thresh
  )
  update_bottlenecks_partial = (
      update_obs_bottlenecks_partial + update_latent_bottlenecks_partial
  )

  latent_bottlenecks_closed = np.sum(latent_sigmas > closed_thresh)
  choice_bottlenecks_closed = np.sum(choice_sigmas > closed_thresh)
  update_obs_bottlenecks_closed = np.sum(update_obs_sigmas > closed_thresh)
  update_latent_bottlenecks_closed = np.sum(
      update_latent_sigmas > closed_thresh
  )
  update_bottlenecks_closed = (
      update_obs_bottlenecks_closed + update_latent_bottlenecks_closed
  )

  bottleneck_dict = {
      'latent_bottlenecks_open': int(latent_bottlenecks_open),
      'latent_bottlenecks_partial': int(latent_bottlenecks_partial),
      'latent_bottlenecks_closed': int(latent_bottlenecks_closed),
      'choice_bottlenecks_open': int(choice_bottlenecks_open),
      'choice_bottlenecks_partial': int(choice_bottlenecks_partial),
      'choice_bottlenecks_closed': int(choice_bottlenecks_closed),
      'update_bottlenecks_open': int(update_bottlenecks_open),
      'update_bottlenecks_partial': int(update_bottlenecks_partial),
      'update_bottlenecks_closed': int(update_bottlenecks_closed),
      }
  return bottleneck_dict


def get_total_sigma(params):
  """Get sum of reparameterized sigmas of a DisRNN."""

  params_disrnn = params['hk_disentangled_rnn']

  latent_bottlenecks = reparameterize_sigma(
      params_disrnn['latent_sigma_params'])
  update_obs_bottlenecks = reparameterize_sigma(
      params_disrnn['update_net_obs_sigma_params'])
  update_latent_bottlenecks = reparameterize_sigma(
      params_disrnn['update_net_latent_sigma_params'])
  choice_bottlenecks = reparameterize_sigma(
      params_disrnn['choice_net_sigma_params'])

  return float(
      jnp.sum(latent_bottlenecks)
      + jnp.sum(update_obs_bottlenecks)
      + jnp.sum(update_latent_bottlenecks)
      + jnp.sum(choice_bottlenecks)
  )


def get_auxiliary_metrics(params: hk.Params) -> dict[str, Any]:
  """Computes auxiliary metrics for the base DisRNN.

  Args:
    params: Haiku parameters of the model.

  Returns:
    A dictionary of auxiliary metrics.
  """

  bottleneck_metrics = log_bottlenecks(params)
  total_sigma_val = get_total_sigma(params)
  return {'total_sigma': total_sigma_val, **bottleneck_metrics}



# Copyright 2025 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for checkpoints."""

from flax import traverse_util
import haiku as hk
import optax


def get_optimizer_with_frozen_params(
    opt: optax.GradientTransformation,
    params: hk.Params,
    trainable_param_names: list[str] | None = None,
) -> optax.GradientTransformation:
  """Modified an optax optimizer to freeze some parameters.

  This function partitions parameters into 'trainable' and 'frozen'.
  'trainable' parameters are optimized using `opt`. 'frozen' parameters
  have their gradients set to zero.

  Args:
    opt: The base Optax optimizer to use for trainable parameters.
    params: The model parameters (e.g., a Haiku parameter dictionary-like
      structure). This is used to identify parameter paths and partition them.
    trainable_param_names: An optional list of strings. If a parameter's path
      (as a tuple of strings) contains any of these strings as a substring in
      any part of the path, it's marked as trainable. If None or empty, all
      parameters will be frozen.

  Returns:
    An Optax `GradientTransformation` that applies `opt` to trainable
    parameters and zeros out gradients for frozen parameters.
  """
  partition_optimizers = {
      'trainable': opt,
      'frozen': optax.set_to_zero(),
  }

  # checks if any of the trainable param names are present in the path as a
  # substring. If so, we mark it as trainable.
  def _is_trainable(path):
    for p in path:
      for trainable_param_name in trainable_param_names:
        if trainable_param_name in p:
          return True
    return False

  # path is a tuple of strings corresponding to the path to each parameter in
  # nested param tree. Example paths:
  # ('hk_disentangled_rnn/choice_mlp', 'input_biases')
  # ('hk_disentangled_rnn', 'dopamine_mlp_multipliers')
  # ('hk_disentangled_rnn/dopamine_mlp', 'input_biases')
  #
  # For each parameter, we check if any of the trainable param names are
  # present in the path. If so, we mark that parameter as trainable.
  # rest of the parameters are frozen.
  param_partitions = traverse_util.path_aware_map(
      lambda path, _: 'trainable' if _is_trainable(path) else 'frozen',
      params,
  )

  traverse_util.path_aware_map(
      lambda path, v: print(path),
      params,
  )
  # Choose the optimizer based on the partition.
  opt = optax.multi_transform(partition_optimizers, param_partitions)
  return opt



# Copyright 2025 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Two armed bandit experiments. Generate synthetic data, plot data."""

import abc
from collections.abc import Callable
from typing import Literal, NamedTuple, Optional, Union
import warnings

from disentangled_rnns.library import rnn_utils
import haiku as hk
import jax
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

abstractmethod = abc.abstractmethod

################
# ENVIRONMENTS #
################


class BaseEnvironment(abc.ABC):
  """Base class for two-armed bandit environments.

  Subclasses must implement the following methods:
    - new_session()
    - step(choice)

  Attributes:
    seed: The seed used to initialize the environment.
    n_arms: The number of arms in the environment.
  """

  def __init__(self, seed: Optional[int] = None, n_arms: int = 2):
    self._random_state = np.random.RandomState(seed)
    self._n_arms = n_arms

  @abstractmethod
  def new_session(self):
    """Starts a new session (e.g., resets environment parameters).

    This method should be implemented by subclasses to initialize or
    reset the environment's state at the beginning of a new session or episode.
    """

  @abstractmethod
  def step(self, attempted_choice: int) -> tuple[int, float, int]:
    """Executes a single step in the environment.

    Args:
      attempted_choice: The action chosen by the agent.

    Returns:
      choice: The action actually taken. May be different from the attempted
        choice if the environment decides the choice should be instructed on
        that trial.
      reward: The reward received after taking the action.
      instructed: 1 if the choice was instructed, 0 otherwise
    """

  @property
  def n_arms(self) -> int:
    """Returns the current reward probabilities for each arm."""
    return self._n_arms


class EnvironmentBanditsDrift(BaseEnvironment):
  """Environment for a drifting two-armed bandit task.

  Reward probabilities on each arm are sampled randomly between 0 and
  1. On each trial, gaussian random noise is added to each, with the constraint
  that the probabilities cannot go below 0 or above 1.

  Attributes:
    sigma: A float, between 0 and 1, giving the magnitude of the drift
    reward_probs: Probability of reward associated with each action
    n_arms: The number of arms in the environment.
  """

  def __init__(self,
               sigma: float,
               p_instructed: float = 0.0,
               seed: Optional[int] = None,
               n_arms: int = 2,
               ):
    super().__init__(seed=seed, n_arms=n_arms)

    # Check inputs
    if sigma < 0:
      msg = ('sigma was {}, but must be greater than 0')
      raise ValueError(msg.format(sigma))

    # Initialize persistent properties
    self._sigma = sigma
    self._p_instructed = p_instructed

    # Sample new reward probabilities
    self.new_session()

  def new_session(self):
    # Pick new reward probabilities.
    # Sample randomly between 0 and 1
    self._reward_probs = self._random_state.rand(self.n_arms)

  def step(self,
           attempted_choice: int) -> tuple[int, float, int]:
    """Run a single trial of the task.

    Args:
      attempted_choice: The choice made by the agent. 0 or 1

    Returns:
      choice: The action actually taken. May be different from the attempted
        choice if the environment decides the choice should be instructed on
        that trial.
      reward: The reward to be given to the agent. 0 or 1.
      instructed: 1 if the choice was instructed, 0 otherwise

    """
    if attempted_choice == -1:
      choice = -1
      reward = -1
      instructed = -1
      return choice, reward, instructed

    # Check inputs
    if attempted_choice not in list(range(self.n_arms)):
      msg = (f'choice given was {attempted_choice}, but must be one of '
             f'{list(range(self.n_arms))}.')
      raise ValueError(msg)

    # If choice was instructed, overrule it and decide randomly
    instructed = self._random_state.rand() < self._p_instructed
    if instructed:
      choice = self._random_state.choice(self.n_arms)
    else:
      choice = attempted_choice

    # Sample reward with the probability of the chosen side
    reward = self._random_state.rand() < self._reward_probs[choice]
    # Add gaussian noise to reward probabilities
    drift = self._random_state.normal(
        loc=0, scale=self._sigma, size=self.n_arms)
    self._reward_probs += drift

    # Fix reward probs that've drifted below 0 or above 1
    self._reward_probs = np.clip(self._reward_probs, 0, 1)

    return choice, float(reward), int(instructed)

  @property
  def reward_probs(self) -> np.ndarray:
    return self._reward_probs.copy()


class NoMoreTrialsInSessionError(ValueError):
  pass


class NoMoreSessionsInDatasetError(ValueError):
  pass


class EnvironmentPayoutMatrix(BaseEnvironment):
  """Environment for a two-armed bandit task with a specified payout matrix."""

  def __init__(
      self,
      payout_matrix: np.ndarray,
      instructed_matrix: Optional[np.ndarray] = None,
  ):
    """Initialize the environment.

    Args:
      payout_matrix: A numpy array of shape (n_sessions, n_actions, n_trials)
        giving the reward for each session, action, and trial. These are
        deterministic, i.e. for the same trial_num, session_num, and action, the
        reward will always be the same. (If you'd like stochastic rewards you
        can populate this matrix ahead of time).
      instructed_matrix: A numpy array of shape (n_sessions, n_trials) giving
        the choice that should be made, if any, for each session and trial.
        Elements should be ints or nan. If nan, the choice is not instructed. If
        None, no choices are instructed.
    """
    n_arms = payout_matrix.shape[2]
    super().__init__(seed=None, n_arms=n_arms)

    self._payout_matrix = payout_matrix
    self._n_sessions = payout_matrix.shape[0]
    self._n_trials = payout_matrix.shape[1]

    if instructed_matrix is not None:
      self._instructed_matrix = instructed_matrix
    else:
      self._instructed_matrix = np.nan * np.zeros_like(payout_matrix)

    self._current_session = 0
    self._current_trial = 0

  def new_session(self):
    self._current_session += 1
    if self._current_session >= self._n_sessions:
      raise NoMoreSessionsInDatasetError(
          'No more sessions in dataset. '
          f'Current session {self._current_session} is out of range '
          f'[0, {self._n_sessions - 1})'
      )
    self._current_trial = 0

  def step(self, attempted_choice: int) -> tuple[int, float, int]:
    # If agent choice is default empty value -1, return -1 for all outputs.
    if attempted_choice == -1:
      choice = -1
      reward = -1
      instructed = -1
      return choice, reward, instructed

    # Check inputted choice is valid.
    if attempted_choice not in list(range(self.n_arms)):
      msg = (f'choice given was {attempted_choice}, but must be one of '
             f'{list(range(self.n_arms))}.')
      raise ValueError(msg)

    if self._current_trial >= self._n_trials:
      raise NoMoreTrialsInSessionError(
          'No more trials in session. '
          f'Current trial {self._current_trial} is out of range '
          f'[0, {self._n_trials})'
      )

    # If choice was instructed, overrule and replace with the instructed choice
    instruction = self._instructed_matrix[
        self._current_session, self._current_trial
    ]
    instructed = not np.isnan(instruction)
    if instructed:
      choice = int(instruction)
    else:
      choice = attempted_choice

    reward = self._payout_matrix[
        self._current_session, self._current_trial, choice
    ]
    self._current_trial += 1
    return choice, float(reward), int(instructed)

  @property
  def payout(self) -> np.ndarray:
    """Get possible payouts for current session, trial across actions."""
    return self._payout_matrix[
        self._current_session, self._current_trial, :].copy()

##########
# AGENTS #
##########


class AgentQ:
  """An agent that runs "vanilla" Q-learning for the y-maze tasks.

  Attributes:
    q: The agent's current estimate of the reward probability on each arm

  """

  def __init__(
      self,
      alpha: float,  # Learning rate
      beta: float,  # softmax temp
  ):
    """Initialize the agent.

    Args:
      alpha: alpha: The agent's learning rate
      beta: The agent's softmax temperature
    """
    self._alpha = alpha
    self._beta = beta
    self.new_session()

  def new_session(self):
    """Reset the agent for the beginning of a new session."""
    self.q = 0.5 * np.ones(2)

  def get_choice_probs(self) -> np.ndarray:
    choice_probs = np.exp(self._beta * self.q) / np.sum(
        np.exp(self._beta * self.q))
    return choice_probs

  def get_choice(self) -> int:
    """Sample a choice, given the agent's current internal state."""

    choice_probs = self.get_choice_probs()
    choice = np.random.choice(2, p=choice_probs)
    return choice

  def update(self,
             choice: int,
             reward: float):
    """Update the agent after one step of the task.

    Args:
      choice: The choice made by the agent. 0 or 1
      reward: The reward received by the agent. 0 or 1
    """
    self.q[choice] = (1 - self._alpha) * self.q[choice] + self._alpha * reward


class AgentLeakyActorCritic:
  """An agent that runs Actor-Critic learning for the y-maze tasks.

  Attributes:
    theta: The actor's state variable. Updated trial-by-trial using a leaky
      gradient bandit algorithm.
    v: The critic's state variable. Updated trial-by-trial to keep a running
      estimate of recent reward rate.
  """

  def __init__(
      self,
      alpha_critic: float,
      alpha_actor_learn: float,
      alpha_actor_forget: float,
  ):
    """Initialize the agent.

    Args:
      alpha_critic: The critic's learning rate
      alpha_actor_learn: The actor's learning rate
      alpha_actor_forget: The actor's forgetting rate
    """
    self._alpha_critic = alpha_critic
    self._alpha_actor_learn = alpha_actor_learn
    self._alpha_actor_forget = alpha_actor_forget
    self.new_session()

  def new_session(self):
    """Reset the agent for the beginning of a new session."""
    self.theta = 0. * np.ones(2)
    self.v = 0.5

  def get_choice_probs(self) -> np.ndarray:
    choice_probs = np.exp(self.theta) / np.sum(
        np.exp(self.theta))
    return choice_probs

  def get_choice(self) -> int:
    """Sample a choice, given the agent's current internal state."""

    choice_probs = self.get_choice_probs()
    choice = np.random.choice(2, p=choice_probs)
    return choice

  def update(self, choice: int, reward: float):
    """Update the agent after one step of the task.

    Args:
      choice: The choice made by the agent. 0 or 1
      reward: The reward received by the agent. 0 or 1
    """
    unchosen = 1 - choice  # Convert 0 to 1 or 1 to 0
    # Actor learning
    # Follows "gradient bandit" algorithm from Sutton & Barto, chapter 2, but
    # with the addition of forgetting
    choice_probs = self.get_choice_probs()
    rpe = reward - self.v
    self.theta[choice] = (1 - self._alpha_actor_forget) * self.theta[
        choice] + self._alpha_actor_learn * rpe * (1 - choice_probs[choice])
    self.theta[unchosen] = (1 - self._alpha_actor_forget) * self.theta[
        unchosen] - self._alpha_actor_learn * rpe * (choice_probs[unchosen])

    # Critic learing: V moves towards reward
    self.v = (1 - self._alpha_critic) * self.v + self._alpha_critic * reward


class AgentNetwork:
  """A class that allows running a trained RNN as an agent.

  Attributes:
    make_network: A Haiku function that returns an RNN architecture
    params: A set of Haiku parameters suitable for that architecture
  """

  def __init__(self,
               make_network: Callable[[], hk.RNNCore],
               params: hk.Params):

    def step_network(
        xs: np.ndarray, state: hk.State
    ) -> tuple[np.ndarray, hk.State]:
      core = make_network()
      y_hat, new_state = core(xs, state)
      return y_hat, new_state

    def get_initial_state() -> hk.State:
      core = make_network()
      rnn_state = core.initial_state(1)
      return rnn_state

    model = hk.without_apply_rng(hk.transform(step_network))
    rnn_state = hk.without_apply_rng(hk.transform(get_initial_state))

    self._initial_state = rnn_state.apply(params)
    self._model_fun = jax.jit(
        lambda xs, state: model.apply(params, xs, rnn_state)
    )
    self._xs = np.zeros((1, 2))
    self.new_session()

  def new_session(self):
    self._rnn_state = self._initial_state

  def get_choice_probs(self) -> np.ndarray:
    output_logits, _ = self._model_fun(self._xs, self._rnn_state)
    choice_probs = np.asarray(jax.nn.softmax(output_logits[0]))
    return choice_probs

  def get_choice(self) -> tuple[int, np.ndarray]:
    choice_probs = self.get_choice_probs()
    choice = np.random.choice(2, p=choice_probs)
    return choice

  def update(self, choice: int, reward: int):
    self._xs = np.array([[choice, reward]])
    _, self._rnn_state = self._model_fun(self._xs, self._rnn_state)


Agent = Union[AgentQ, AgentLeakyActorCritic, AgentNetwork]


class SessData(NamedTuple):
  choices: np.ndarray
  rewards: np.ndarray
  reward_probs: np.ndarray
  n_trials: int


def run_experiment(agent: Agent,
                   environment: EnvironmentBanditsDrift,
                   n_steps: int) -> SessData:
  """Runs a behavioral session from a given agent and environment.

  Args:
    agent: An agent object
    environment: An environment object
    n_steps: The number of steps in the session you'd like to generate

  Returns:
    experiment: A SessData object holding choices and rewards from the session
  """
  choices = np.zeros(n_steps)
  rewards = np.zeros(n_steps)
  reward_probs = np.zeros((n_steps, 2))

  for step in np.arange(n_steps):
    # First record environment reward probs
    reward_probs[step] = environment.reward_probs
    # First agent makes a choice
    attempted_choice = agent.get_choice()
    # Then environment computes a reward
    choice, reward, _ = environment.step(attempted_choice)
    # Finally agent learns
    agent.update(choice, reward)
    # Log choice and reward
    choices[step] = choice
    rewards[step] = reward

  experiment = SessData(choices=choices,
                        rewards=rewards,
                        n_trials=n_steps,
                        reward_probs=reward_probs)
  return experiment


def create_dataset(agent: Agent,
                   environment: EnvironmentBanditsDrift,
                   n_steps_per_session: int,
                   n_sessions: int,
                   batch_size: int) -> rnn_utils.DatasetRNN:
  """Generates a behavioral dataset from a given agent and environment.

  Args:
    agent: An agent object to generate choices
    environment: An environment object to generate rewards
    n_steps_per_session: The number of trials in each behavioral session to
      be generated
    n_sessions: The number of sessions to generate
    batch_size: The size of the batches to serve from the dataset

  Returns:
    rnn_utils.DatasetRNN object
  """
  xs = np.zeros((n_steps_per_session, n_sessions, 2))
  ys = np.zeros((n_steps_per_session, n_sessions, 1))

  for sess_i in np.arange(n_sessions):
    environment.new_session()
    agent.new_session()
    experiment = run_experiment(agent, environment, n_steps_per_session)
    prev_choices = np.concatenate(([0], experiment.choices[0:-1]))
    prev_rewards = np.concatenate(([0], experiment.rewards[0:-1]))
    xs[:, sess_i] = np.swapaxes(
        np.concatenate(([prev_choices], [prev_rewards]), axis=0), 0, 1
    )
    ys[:, sess_i] = np.expand_dims(experiment.choices, 1)

  dataset = rnn_utils.DatasetRNN(
      xs=xs,
      ys=ys,
      x_names=['prev choice', 'prev reward'],
      y_names=['choice'],
      y_type='categorical',
      n_classes=2,
      batch_size=batch_size,
  )
  return dataset


def plot_2ab_sessdata(
    choices: np.ndarray,
    rewards: np.ndarray,
    scalars: np.ndarray | None = None,
    scalar_types: Literal['reward_probs', 'agent_states', 'other'] = 'other',
    show_legend: bool = True,
    left_color: str = 'rebeccapurple',
    right_color: str = 'darkorange',
):
  """Creates a figure showing data from session of a two-armed binary bandit task.

  Args:
    choices: The choices made by the agent in the session
    rewards: The rewards received by the agent in the session
    scalars: Any scalars output by the agent in the session
    scalar_types: Whether the scalars are reward probabilities or agent states
    show_legend: Whether to show the legend of the scalars.
    left_color: The color to use for the "left" arm (choice 0).
    right_color: The color to use for the "right" arm (choice 1).
  """

  # Check that choices and rewards are both binary. Valid values are -1, 0,
  # and 1
  if not np.all(np.isin(choices, [-1, 0, 1])):
    warnings.warn(
        'Choices should be either -1, 0, or 1, but got choices: '
        f'{np.unique(choices)}.'
    )
  if not np.all(np.isin(rewards, [-1, 0, 1])):
    warnings.warn(
        'Rewards should be either -1, 0, or 1, but got rewards: '
        f'{np.unique(rewards)}.'
    )

  choose_high = choices == 1
  choose_low = choices == 0
  rewarded = rewards == 1

  # Get the RdYlGn colormap
  cmap = plt.cm.RdYlGn
  # Get the RGBA values for reward (green) and omission (red) ticks
  omission_color = cmap(0.1)  # A red that's near the bottom of the colormap
  reward_color = cmap(0.9)  # A green that's near the top of the colormap

  # Check whether we have scalars we'd like to plot.
  if scalars is None:
    scalar_names = []
    scalars = []
    ymax = 1
    ymin = 0
    scalar_colors = []
    ylabel = ''
  elif scalar_types == 'reward_probs':
    scalar_names = ['p(reward|left)', 'p(reward|right)']
    scalar_colors = [left_color, right_color]
    ymax = 1
    ymin = 0
    ylabel = 'Reward Probability'
  elif scalar_types == 'agent_states':
    if scalars.ndim == 1:
      scalars = np.expand_dims(scalars, 1)
    scalar_names = [f'state {i}' for i in range(scalars.shape[1])]
    scalar_colors = sns.color_palette('colorblind')
    ymax = np.max(scalars)
    ymin = np.min(scalars)
    ylabel = 'Agent State'
  elif scalar_types == 'other':
    scalar_names = []
    scalars = []
    ymax = np.max(scalars)
    ymin = np.min(scalars)
    scalar_colors = []
    ylabel = ''
  else:
    raise ValueError(
        'scalar_types must be one of "reward_probs", "agent_states", or '
        f'"other", but got {scalar_types}'
    )

  # Make the plot
  plt.subplots(figsize=(10, 3))
  # Plot the scalars
  for scalar_i in range(len(scalar_names)):
    plt.plot(
        scalars[:, scalar_i], color=scalar_colors[scalar_i % len(scalar_colors)]
    )
  if show_legend:
    plt.legend(scalar_names)

  # Rewarded high: Long green ticks on top
  top = ymax + 0.1 * (ymax - ymin)
  bottom = ymin - 0.1 * (ymax - ymin)
  plt.scatter(
      np.argwhere(choose_high & rewarded),
      np.ones(np.sum(choose_high & rewarded)) * top,
      color=reward_color,
      marker='|',
  )

  # Omission high: Short red ticks on top
  plt.scatter(
      np.argwhere(choose_high & 1 - rewarded),
      np.ones(np.sum(choose_high & 1 - rewarded)) * top,
      color=omission_color,
      marker='|',
  )

  # Rewarded low: Green ticks on bottom
  plt.scatter(
      np.argwhere(choose_low & rewarded),
      np.ones(np.sum(choose_low & rewarded)) * bottom,
      color=reward_color,
      marker='|',
  )

  # Omission Low: Red ticks on bottom
  plt.scatter(
      np.argwhere(choose_low & 1 - rewarded),
      np.ones(np.sum(choose_low & 1 - rewarded)) * bottom,
      color=omission_color,
      marker='|',
  )

  plt.xlabel('Trial')
  plt.xlim([0, len(choices)])
  plt.ylabel(ylabel)
  if ylabel == 'Reward Probability':
    plt.yticks([0, 0.5, 1], ['0%', '50%', '100%'])



# Copyright 2025 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Plotting functions for inspecting Disentangled RNNs."""

import copy
from typing import Optional

from disentangled_rnns.library import disrnn
from disentangled_rnns.library import multisubject_disrnn
from disentangled_rnns.library import rnn_utils
import haiku as hk
import jax
import jax.numpy as jnp
import matplotlib as mpl
from matplotlib import pyplot as plt
import numpy as np


# Fontsizes and formatting for plots.
small = 15
medium = 18
large = 20
mpl.rcParams['grid.color'] = 'none'
mpl.rcParams['axes.facecolor'] = 'white'
plt.rcParams['svg.fonttype'] = 'none'


def plot_bottlenecks(
    params: hk.Params,
    disrnn_config: disrnn.DisRnnConfig,
    sort_latents: bool = True,
) -> plt.Figure:
  """Plot the bottleneck sigmas from an hk.DisentangledRNN."""

  if isinstance(disrnn_config, multisubject_disrnn.MultisubjectDisRnnConfig):
    params_disrnn = params['multisubject_dis_rnn']
    subject_embedding_size = disrnn_config.subject_embedding_size
    update_input_names = [
        f'SubjEmb {i+1}' for i in range(subject_embedding_size)
    ] + disrnn_config.x_names[1:]
    # For update_sigmas: concatenate transposed reparameterized sigmas
    # Order of inputs to update nets: subject_embedding, observations, latents
    update_subj_sigmas_t = np.transpose(
        disrnn.reparameterize_sigma(
            params_disrnn['update_net_subj_sigma_params']
        )
    )
    update_obs_sigmas_t = np.transpose(
        disrnn.reparameterize_sigma(
            params_disrnn['update_net_obs_sigma_params']
        )
    )
    update_latent_sigmas_t = np.transpose(
        disrnn.reparameterize_sigma(
            params_disrnn['update_net_latent_sigma_params']
        )
    )
    update_sigmas = np.concatenate(
        (update_subj_sigmas_t, update_obs_sigmas_t, update_latent_sigmas_t),
        axis=1,
    )

    # For choice_sigmas: concatenate reparameterized sigmas
    # Order of inputs to choice net: subject_embedding, latents
    choice_subj_sigmas = disrnn.reparameterize_sigma(
        params_disrnn['choice_net_subj_sigma_params']
    )
    choice_latent_sigmas = disrnn.reparameterize_sigma(
        params_disrnn['choice_net_latent_sigma_params']
    )
    choice_sigmas = np.concatenate((choice_subj_sigmas, choice_latent_sigmas))
  elif isinstance(disrnn_config, disrnn.DisRnnConfig):
    params_disrnn = params['hk_disentangled_rnn']
    subject_embedding_size = 0
    update_input_names = disrnn_config.x_names
    # For update_sigmas: concatenate transposed reparameterized sigmas
    # Order of inputs to update nets: observations, latents
    update_obs_sigmas_t = np.transpose(
        disrnn.reparameterize_sigma(
            params_disrnn['update_net_obs_sigma_params']
        )
    )
    update_latent_sigmas_t = np.transpose(
        disrnn.reparameterize_sigma(
            params_disrnn['update_net_latent_sigma_params']
        )
    )
    update_sigmas = np.concatenate(
        (update_obs_sigmas_t, update_latent_sigmas_t), axis=1)
    choice_sigmas = np.array(
        disrnn.reparameterize_sigma(
            np.transpose(params_disrnn['choice_net_sigma_params'])
        )
    )
  else:
    raise ValueError(
        'plot_bottlenecks only supports DisRnnConfig and'
        ' MultisubjectDisRnnConfig.'
    )

  latent_sigmas = np.array(
      disrnn.reparameterize_sigma(params_disrnn['latent_sigma_params'])
  )

  if sort_latents:
    latent_sigma_order = np.argsort(latent_sigmas)
    latent_sigmas = latent_sigmas[latent_sigma_order]

    # Sort choice sigmas based on the order of latents, keeping subject
    # embedding dimensions first if they exist.
    choice_sigma_order = np.concatenate(
        (
            np.arange(0, subject_embedding_size),
            subject_embedding_size + latent_sigma_order,
        ),
        axis=0,
    )
    choice_sigmas = choice_sigmas[choice_sigma_order]

    # Sort update sigmas based on the order of latents, keeping subject
    # embedding dimensions first if they exist, then observations, then latents.
    non_latent_input_size = subject_embedding_size + disrnn_config.obs_size
    update_sigma_order = np.concatenate(
        (
            np.arange(0, non_latent_input_size, 1),
            non_latent_input_size + latent_sigma_order,
        ),
        axis=0,
    )
    update_sigmas = update_sigmas[latent_sigma_order, :]
    update_sigmas = update_sigmas[:, update_sigma_order]

  latent_names = np.arange(1, disrnn_config.latent_size + 1)
  fig, axes = plt.subplots(1, 3, figsize=(15, 5))

  # Plot Latent Bottlenecks on axes[0]
  im1 = axes[0].imshow(np.swapaxes([1 - latent_sigmas], 0, 1), cmap='Oranges')
  im1.set_clim(vmin=0, vmax=1)
  axes[0].set_yticks(
      ticks=range(disrnn_config.latent_size),
      labels=latent_names,
      fontsize=small,
  )
  axes[0].set_xticks(ticks=[])
  axes[0].set_ylabel('Latent #', fontsize=medium)
  axes[0].set_title('Latent Bottlenecks', fontsize=large)

  # Plot Choice Bottlenecks on axes[1]
  # These bottlenecks apply to the inputs of the choice network:
  # [subject embeddings, latents]
  choice_input_dim = subject_embedding_size + disrnn_config.latent_size
  choice_input_names = np.concatenate((
      [f'SubjEmb {i+1}' for i in range(subject_embedding_size)],
      [f'Latent {i}' for i in latent_names]
  ))
  im2 = axes[1].imshow(np.swapaxes([1 - choice_sigmas], 0, 1), cmap='Oranges')
  im2.set_clim(vmin=0, vmax=1)
  axes[1].set_yticks(
      ticks=range(choice_input_dim), labels=choice_input_names, fontsize=small
  )
  axes[1].set_xticks(ticks=[])
  axes[1].set_ylabel('Choice Network Input', fontsize=medium)
  axes[1].set_title('Choice Network Bottlenecks', fontsize=large)

  # Plot Update Bottlenecks on axes[2]
  im3 = axes[2].imshow(1 - update_sigmas, cmap='Oranges')
  im3.set_clim(vmin=0, vmax=1)
  cbar = fig.colorbar(im3, ax=axes[2])
  # Y-axis corresponds to the target latent (sorted if sort_latents=True)
  cbar.ax.tick_params(labelsize=small)
  axes[2].set_yticks(
      ticks=range(disrnn_config.latent_size),
      labels=latent_names,
      fontsize=small,
  )
  # X-axis corresponds to the inputs to the update network:
  # [subject embeddings, observations, latents]
  xlabels = update_input_names + [f'Latent {i}' for i in latent_names]
  axes[2].set_xticks(
      ticks=range(len(xlabels)),
      labels=xlabels,
      rotation='vertical',
      fontsize=small,
  )
  axes[2].set_ylabel('Latent #', fontsize=medium)
  axes[2].set_xlabel('Update Network Inputs', fontsize=medium)
  axes[2].set_title('Update Network Bottlenecks', fontsize=large)
  fig.tight_layout()  # Adjust layout to prevent overlap
  return fig


def plot_update_rules(
    params: hk.Params,
    disrnn_config: disrnn.DisRnnConfig,
    subj_ind: Optional[int] = None,
    axis_lim: float = 2.1,
) -> list[plt.Figure]:
  """Generates visualizations of the update rules of a HkDisentangledRNN."""

  disrnn_config = copy.deepcopy(disrnn_config)
  disrnn_config.noiseless_mode = True  # Turn off noise for plotting

  if isinstance(disrnn_config, multisubject_disrnn.MultisubjectDisRnnConfig):
    if subj_ind is None:
      print('In multisubject mode, but subj_ind not provided. Defaulting to 0')
      subj_ind = 0
  elif subj_ind is not None:
    print('subj_ind provided, but not in multisubject mode. Ignoring it')
    subj_ind = None

  if isinstance(disrnn_config, multisubject_disrnn.MultisubjectDisRnnConfig):
    make_network = lambda: multisubject_disrnn.MultisubjectDisRnn(disrnn_config)
    obs_names = disrnn_config.x_names[1:]  # First x_name is "Subject ID"
    param_prefix = 'multisubject_dis_rnn'
    subj_embedding_size = disrnn_config.subject_embedding_size
    update_subj_s_t = np.transpose(disrnn.reparameterize_sigma(
        params[param_prefix]['update_net_subj_sigma_params']))
    update_obs_s_t = np.transpose(disrnn.reparameterize_sigma(
        params[param_prefix]['update_net_obs_sigma_params']))
    update_latent_s_t = np.transpose(disrnn.reparameterize_sigma(
        params[param_prefix]['update_net_latent_sigma_params']))
    update_sigmas = np.concatenate(
        (update_subj_s_t, update_obs_s_t, update_latent_s_t), axis=1
    )
  elif isinstance(disrnn_config, disrnn.DisRnnConfig):
    make_network = lambda: disrnn.HkDisentangledRNN(disrnn_config)
    obs_names = disrnn_config.x_names
    param_prefix = 'hk_disentangled_rnn'
    subj_embedding_size = 0
    update_obs_s_t = np.transpose(
        disrnn.reparameterize_sigma(
            params[param_prefix]['update_net_obs_sigma_params']
        )
    )
    update_latent_s_t = np.transpose(
        disrnn.reparameterize_sigma(
            params[param_prefix]['update_net_latent_sigma_params']
        ))
    update_sigmas = np.concatenate((update_obs_s_t, update_latent_s_t), axis=1)

  else:
    raise ValueError(
        f'Unsupported config type: {type(disrnn_config)} for plot_update_rules.'
    )

  def step(xs, state):
    core = make_network()
    output, new_state = core(jnp.expand_dims(jnp.array(xs), axis=0), state)
    return output, new_state

  _, step_hk = hk.transform(step)
  key = jax.random.PRNGKey(0)
  step_hk = jax.jit(step_hk)

  initial_state = np.array(rnn_utils.get_initial_state(make_network))
  reference_state = np.zeros(initial_state.shape)

  def plot_update_1d(params, unit_i, observations, titles):
    state_bins = np.linspace(-axis_lim, axis_lim, 20)
    colormap = mpl.colormaps['viridis'].resampled(3)
    colors = colormap.colors

    fig, axes = plt.subplots(
        1, len(observations), figsize=(len(observations) * 4, 5.5), sharey=True
    )
    # Ensure axes is always an array for consistent indexing
    if len(observations) == 1:
      axes = [axes]
    axes[0].set_ylabel('Δ Activity')

    for observation_i in range(len(observations)):
      observation = observations[observation_i]
      if subj_ind is not None:
        observation = [subj_ind] + observation
      ax = axes[observation_i]
      delta_states = np.zeros(shape=(len(state_bins), 1))
      for s_i in np.arange(len(state_bins)):
        state = reference_state
        state[0, unit_i] = state_bins[s_i]
        _, next_state = step_hk(
            params, key, observation, state
        )
        next_state = np.array(next_state)
        delta_states[s_i] = next_state[0, unit_i] - state_bins[s_i]

      ax.plot((-axis_lim, axis_lim), (0, 0), color='black')
      ax.plot(state_bins, delta_states, color=colors[1])
      ax.set_title(titles[observation_i], fontsize=large)
      ax.set_xlim(-axis_lim, axis_lim)
      ax.set_xlabel(
          'Latent ' + str(unit_i + 1) + ' Activity', fontsize=medium
      )
      ax.set_aspect('equal')
      ax.tick_params(axis='both', labelsize=small)

    return fig

  def plot_update_2d(params, unit_i, unit_input, observations, titles):

    state_bins = np.linspace(-axis_lim, axis_lim, 50)
    state_bins_input = np.linspace(-axis_lim/2, axis_lim/2, 5)
    colormap = mpl.colormaps['viridis'].resampled(len(state_bins_input))
    colors = colormap.colors

    fig, axes = plt.subplots(
        1,
        len(observations),
        figsize=(len(observations) * 2 + 10, 5.5),
        sharey=True,
    )
    # Ensure axes is always an array for consistent indexing
    if len(observations) == 1:
      axes = [axes]
    axes[0].set_ylabel('Δ Activity', fontsize=medium)

    for observation_i in range(len(observations)):
      observation = observations[observation_i]
      if subj_ind is not None:
        observation = [subj_ind] + observation
      legend_elements = []
      ax = axes[observation_i]
      for si_i in np.arange(len(state_bins_input)):
        delta_states = np.zeros(shape=(len(state_bins), 1))
        for s_i in np.arange(len(state_bins)):
          state = reference_state
          state[0, unit_i] = state_bins[s_i]
          state[0, unit_input] = state_bins_input[si_i]

          _, next_state = step_hk(params, key, observation, state)
          next_state = np.array(next_state)
          delta_states[s_i] = next_state[0, unit_i] - state_bins[s_i]

        lines = ax.plot(state_bins, delta_states, color=colors[si_i])
        legend_elements.append(lines[0])

        if observation_i == 0:
          legend_labels = [f'{num:.1f}' for num in state_bins_input]  # pylint: disable=bad-whitespace
          ax.legend(legend_elements, legend_labels, fontsize=small)

      ax.plot((-axis_lim, axis_lim), (0, 0), color='black')
      ax.set_title(titles[observation_i], fontsize=large)
      ax.set_xlim(-axis_lim, axis_lim)
      ax.set_xlabel(
          'Latent ' + str(unit_i + 1) + ' Activity', fontsize=medium
      )
      ax.tick_params(axis='both', labelsize=small)

    return fig

  latent_sigmas = np.array(
      disrnn.reparameterize_sigma(
          params[param_prefix]['latent_sigma_params']
      )
  )

  # TODO(kevinjmiller): Generalize to allow different observation length
  if disrnn_config.obs_size != 2:
    raise NotImplementedError(
        'Plot update rules currently assumes that there are exactly two'
        f' observations. Instead founc observarions {obs_names}'
    )

  latent_order = np.argsort(latent_sigmas)
  figs = []

  # Loop over latents. Plot update rules
  for latent_i in latent_order:
    # If this latent's bottleneck is open
    if latent_sigmas[latent_i] < 0.5:

      # Which of its input bottlenecks are open?
      update_net_inputs = np.argwhere(update_sigmas[latent_i] < 0.5)
      obs1_sensitive = np.any(update_net_inputs == subj_embedding_size)
      obs2_sensitive = np.any(update_net_inputs == subj_embedding_size + 1)

      # Choose which observations to use based on input bottlenecks
      if obs1_sensitive and obs2_sensitive:
        observations = ([0, 0], [0, 1], [1, 0], [1, 1])
        titles = (
            obs_names[0] + ': 0\n' + obs_names[1] + ': 0',
            obs_names[0] + ': 0\n' + obs_names[1] + ': 1',
            obs_names[0] + ': 1\n' + obs_names[1] + ': 0',
            obs_names[0] + ': 1\n' + obs_names[1] + ': 1',
        )
      elif obs1_sensitive:
        observations = ([0, 0], [1, 0])
        titles = (obs_names[0] + ': 0', obs_names[0] + ': 1')
      elif obs2_sensitive:
        observations = ([0, 0], [0, 1])
        titles = (obs_names[1] + ': 0', obs_names[1] + ': 1')
      else:
        observations = ([0, 0],)
        titles = ('All Trials',)

      # Choose which other latents to condition on, based on input bottlenecks
      start_idx_of_latents = subj_embedding_size + disrnn_config.obs_size
      is_latent_input_mask = update_net_inputs >= start_idx_of_latents
      influential_latent_input_indices = update_net_inputs[is_latent_input_mask]
      update_net_input_latents = (
          influential_latent_input_indices - start_idx_of_latents
      )
      # Doesn't count if it depends on itself (this'll be shown no matter what)
      latent_sensitive = np.delete(
          update_net_input_latents, update_net_input_latents == latent_i
      )
      if not latent_sensitive.size:  # Depends on no other latents
        fig = plot_update_1d(params, latent_i, observations, titles)
      else:  # It depends on latents other than itself.
        fig = plot_update_2d(
            params,
            latent_i,
            latent_sensitive[0],
            observations,
            titles,
        )
      if len(latent_sensitive) > 1:
        print(
            'WARNING: This update rule depends on more than one '
            'other latent. Plotting just one of them'
        )
      figs.append(fig)
      fig.tight_layout()

  return figs


def plot_choice_rule(
    params: hk.Params,
    disrnn_config: disrnn.DisRnnConfig,
    subj_embedding: Optional[np.ndarray] = None,
    axis_lim: float = 2.1,
) -> Optional[plt.Figure]:
  """Plots the choice rule of a DisRNN.

  Args:
    params: The parameters of the DisRNN
    disrnn_config: A DisRnnConfig object
    subj_embedding: The subject embedding to use. If None, use a zero vector
      (loosely: the average subject)
    axis_lim: The axis limit for the plot.

  Returns:
    A matplotlib Figure object, or None if choice depends on no latents.
  """

  disrnn_config = copy.deepcopy(disrnn_config)
  disrnn_config.noiseless_mode = True  # Turn off noise for plotting
  activation_fn = getattr(jax.nn, disrnn_config.activation)

  if isinstance(disrnn_config, multisubject_disrnn.MultisubjectDisRnnConfig):
    subj_embedding_size = disrnn_config.subject_embedding_size
    params_prefix = 'multisubject_dis_rnn'
    choice_subj_s = disrnn.reparameterize_sigma(
        params[params_prefix]['choice_net_subj_sigma_params']
    )
    choice_latent_s = disrnn.reparameterize_sigma(
        params[params_prefix]['choice_net_latent_sigma_params']
    )
    choice_net_sigmas = np.concatenate((choice_subj_s, choice_latent_s))
  elif isinstance(disrnn_config, disrnn.DisRnnConfig):
    subj_embedding_size = 0
    params_prefix = 'hk_disentangled_rnn'
    choice_net_sigmas = disrnn.reparameterize_sigma(
        params[params_prefix]['choice_net_sigma_params']
    )
  else:
    raise ValueError(
        'DisRnnConfig is neither MultisubjectDisRnnConfig nor DisRnnConfig,'
        f' but got {type(disrnn_config).__name__}'
    )

  if subj_embedding is None:
    subj_embedding = np.zeros(shape=(subj_embedding_size,))

  n_vals = 100

  def forward(xs):
    choice_net_output = disrnn.ResMLP(
        input_size=disrnn_config.latent_size + subj_embedding_size,
        output_size=disrnn_config.output_size,
        n_units_per_layer=disrnn_config.choice_net_n_units_per_layer,
        n_layers=disrnn_config.choice_net_n_layers,
        activation_fn=activation_fn,
        name='choice_net',
    )(xs)
    return choice_net_output

  model = hk.transform(forward)
  apply = jax.jit(model.apply)

  choice_net_params = {
      'choice_net': params[params_prefix + '/~predict_targets/choice_net']
  }

  # Determine which latents to vary based on their choice_net_sigma_params.
  # choice_net_sigmas has shape (subj_embedding_size + latent_size,).
  latent_to_choice_net_sigmas = choice_net_sigmas[subj_embedding_size:]

  sorted_latent_indices = np.argsort(
      latent_to_choice_net_sigmas
  )

  influential_latents_indices_in_latent_space = [
      latent_idx  # This is an index in the latent space (0 to latent_size-1)
      for latent_idx in sorted_latent_indices
      if latent_to_choice_net_sigmas[latent_idx] < 0.5
  ]
  n_latents_to_plot = min(len(influential_latents_indices_in_latent_space), 2)

  if n_latents_to_plot == 0:
    print(
        'Choice rule: No latents have a choice_net_input_sigma < 0.5.'
        ' Plotting not possible.'
    )
    return None

  # Select the actual latents to vary (indices within the latent space)
  varying_latents_plot_indices = influential_latents_indices_in_latent_space[
      :n_latents_to_plot
  ]

  if n_latents_to_plot == 1:
    # Choice Rule 1D: A curve
    policy_latent_idx_in_latent_space = varying_latents_plot_indices[0]
    policy_latent_vals = np.linspace(-axis_lim, axis_lim, n_vals)
    xs = np.zeros((
        n_vals,
        subj_embedding_size + disrnn_config.latent_size,
    ))
    xs[:, :subj_embedding_size] = subj_embedding
    # Vary the selected latent; other latents remain 0
    xs[
        :, subj_embedding_size + policy_latent_idx_in_latent_space
    ] = policy_latent_vals
    choice_net_output = apply(choice_net_params, jax.random.PRNGKey(0), xs)
    y_hats = choice_net_output[0]
    choice_logits = y_hats[:, 1] - y_hats[:, 0]

    fig, ax = plt.subplots()
    ax.plot(policy_latent_vals, choice_logits, 'g')
    ax.set_title('Choice Rule', fontsize=large)
    ax.set_xlabel(
        f'Latent {policy_latent_idx_in_latent_space + 1}', fontsize=medium
    )
    ax.set_ylabel('Choice Logit', fontsize=medium)
    ax.tick_params(axis='both', labelsize=small)

  else:
    # Choice Rule 2D: A colormap
    if len(influential_latents_indices_in_latent_space) > 2:
      print(
          'WARNING: More than two latents have choice_net_input_sigma < 0.5.'
          ' Plotting only the two with the smallest choice_net_input_sigmas.'
      )

    policy_latent_idx1_in_latent_space = varying_latents_plot_indices[0]
    policy_latent_idx2_in_latent_space = varying_latents_plot_indices[1]

    latent_vals = np.linspace(-axis_lim, axis_lim, n_vals)

    xv, yv = np.meshgrid(latent_vals, latent_vals)
    latent0_vals = xv.flatten()
    latent1_vals = yv.flatten()

    xs = np.zeros((
        n_vals * n_vals,
        subj_embedding_size + disrnn_config.latent_size,
    ))
    xs[:, :subj_embedding_size] = subj_embedding
    # Vary the selected latents; other latents remain 0
    xs[
        :, subj_embedding_size + policy_latent_idx1_in_latent_space
    ] = latent0_vals
    xs[
        :, subj_embedding_size + policy_latent_idx2_in_latent_space
    ] = latent1_vals

    y_hats = apply(choice_net_params, jax.random.PRNGKey(0), xs)
    # TODO(kevinjmiller): This assumes two-alternative logits. Generalize to
    # K-alternative choice. For now, take difference between first two.
    choice_logits_2d = y_hats[0][:, 1] - y_hats[0][:, 0]
    choice_logits_2d = choice_logits_2d.reshape((n_vals, n_vals))

    fig, ax = plt.subplots()
    scatter = ax.imshow(
        choice_logits_2d,
        cmap='coolwarm',
        origin='lower',
        extent=[-axis_lim, axis_lim, -axis_lim, axis_lim],
        aspect='auto',
    )
    cbar = fig.colorbar(scatter, ax=ax)
    cbar.ax.tick_params(labelsize=small)
    cbar.set_label('Choice Logit', fontsize=medium)
    ax.set_title('Choice Rule', fontsize=large)
    ax.set_xlabel(
        f'Latent {policy_latent_idx1_in_latent_space + 1}', fontsize=medium
    )
    ax.set_ylabel(
        f'Latent {policy_latent_idx2_in_latent_space + 1}', fontsize=medium
    )
    ax.tick_params(axis='both', labelsize=small)

  return fig



# Copyright 2025 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Syntehtic two armed bandit datasets with simulated dopamine."""

from typing import NamedTuple, Union

from disentangled_rnns.library import rnn_utils
from disentangled_rnns.library import two_armed_bandits
import numpy as np


class AgentRewardSeeking:
  """An agent that uses a single 'reward-seeking' variable.

  Computes prediction error as the difference between reward and this variable.
  """

  def __init__(
      self,
      alpha: float,  # Learning rate
      beta: float,  # softmax temp
  ):
    """Initialize the agent.

    Args:
      alpha: alpha: The agent's learning rate
      beta: The agent's softmax temperature
    """
    self._alpha_r = alpha
    self._beta_r = beta
    self.new_session()

  def new_session(self):
    """Reset the agent for the beginning of a new session."""
    self.r = 0.0
    self.last_rpe = 0.0

  def get_choice_probs(self) -> np.ndarray:

    choice_logits = np.array([-1, 1]) * (self._beta_r * self.r)
    choice_probs = np.exp(choice_logits) / np.sum(np.exp(choice_logits))
    return choice_probs

  def get_choice(self) -> int:
    """Sample a choice, given the agent's current internal state."""

    choice_probs = self.get_choice_probs()
    choice = np.random.choice(2, p=choice_probs)
    return choice

  def get_rpe(self) -> float:
    return self.last_rpe

  def update(self, choice: int, reward: int):
    """Update the agent after one step of the task.

    Args:
      choice: The choice made by the agent. 0 or 1
      reward: The reward received by the agent. 0 or 1
    """
    assert choice in [0, 1]
    assert reward in [0, 1]
    choice_for_update = 2 * choice - 1  # Convert 0, 1 or -1, 1
    reward_for_update = 2 * reward - 1  # Convert 0, 1 to -1,1

    # Since 'r' is a sided policy variable, the proxy for 'reward
    # expectation' is choice_for_update * r
    self.last_rpe = reward_for_update - choice_for_update * self.r

    # To update the sided variable we need to use a signed RPE
    self.r = self.r + self._alpha_r * choice_for_update * self.last_rpe


class AgentQ(two_armed_bandits.AgentQ):
  """An agent that runs "vanilla" Q-learning for the y-maze tasks.

  Also keeps track of the reward prediction error (RPE).
  """

  def __init__(
      self,
      alpha: float,  # Learning rate
      beta: float,  # softmax temp
  ):
    super().__init__(alpha, beta)
    self.last_rpe = 0.0

  def get_rpe(self) -> float:
    return self.last_rpe

  def update(self, choice: int, reward: float):
    """Update the agent after one step of the task."""
    # First, we calculate the RPE as the difference between the actual reward
    # and the predicted reward.
    self.last_rpe = reward - self.q[choice]
    # Now we update the agent's Q-values (i.e its predictions for rewards).
    super().update(choice, reward)


class SessData(NamedTuple):
  choices: np.ndarray
  rewards: np.ndarray
  reward_probs: np.ndarray
  rpes: np.ndarray
  n_trials: int


Agent = Union[AgentQ, AgentRewardSeeking]


def run_experiment(
    agent: Agent,
    environment: two_armed_bandits.EnvironmentBanditsDrift,
    n_steps: int,
) -> SessData:
  """Runs a behavioral session from a given agent and environment."""
  choices = np.zeros(n_steps)
  rewards = np.zeros(n_steps)
  reward_probs = np.zeros((n_steps, 2))
  rpes = np.zeros(n_steps)

  for step in np.arange(n_steps):
    # First record environment reward probs
    reward_probs[step] = environment.reward_probs

    attempted_choice = agent.get_choice()
    choice, reward, _ = environment.step(attempted_choice)
    agent.update(choice, reward)
    rpe = agent.get_rpe()

    choices[step] = choice
    rewards[step] = reward
    rpes[step] = rpe

  experiment = SessData(
      choices=choices,
      rewards=rewards,
      n_trials=n_steps,
      rpes=rpes,
      reward_probs=reward_probs,
  )
  return experiment


def create_dataset(
    agent: Agent,
    environment: two_armed_bandits.EnvironmentBanditsDrift,
    n_steps_per_session: int,
    n_sessions: int,
    batch_size: int,
) -> rnn_utils.DatasetRNN:
  """Generates a behavioral dataset from a given agent and environment.

  Args:
    agent: An agent object to generate choices and RPEs. Note that this agent
      must be ones of the agents described in this class, i.e. something that
      generates the reward prediction error (RPE) as well.
    environment: An environment object to generate rewards
    n_steps_per_session: The number of trials in each behavioral session to be
      generated
    n_sessions: The number of sessions to generate
    batch_size: The size of the batches to serve from the dataset

  Returns:
  """
  xs = np.zeros((n_steps_per_session, n_sessions, 2))
  ys = np.zeros((n_steps_per_session, n_sessions, 2))

  for sess_i in np.arange(n_sessions):
    environment.new_session()
    agent.new_session()
    experiment = run_experiment(agent, environment, n_steps_per_session)
    prev_choices = np.concatenate(([0], experiment.choices[0:-1]))
    prev_rewards = np.concatenate(([0], experiment.rewards[0:-1]))
    prev_rpes = np.concatenate(([0], experiment.rpes[0:-1]))
    xs[:, sess_i] = np.swapaxes(
        np.concatenate(([prev_choices], [prev_rewards]), axis=0), 0, 1
    )
    ys[:, sess_i] = np.swapaxes(
        np.concatenate(([experiment.choices], [prev_rpes]), axis=0), 0, 1
    )

  dataset = rnn_utils.DatasetRNN(
      xs=xs,
      ys=ys,
      x_names=['prev choice', 'prev reward'],
      y_names=['choice', 'dopamine'],
      y_type='mixed',
      n_classes=2,
      batch_size=batch_size,
  )
  return dataset


def get_q_learning_with_dopamine_dataset(
    n_sessions: int = 500,
    alpha: float = 0.3,
    beta: float = 3,
    n_trials: int = 500,
) -> rnn_utils.DatasetRNN:
  """Generates synthetic dataset of behavior and dopamine from Q-Learning agent."""
  np.random.seed(0)
  agent = AgentQ(alpha=alpha, beta=beta)
  environment = two_armed_bandits.EnvironmentBanditsDrift(sigma=0.1)
  dataset = create_dataset(
      agent,
      environment,
      n_steps_per_session=n_trials,
      n_sessions=n_sessions,
      batch_size=n_sessions,
  )
  return dataset


def get_reward_seeking_with_dopamine_dataset(
    n_sessions: int = 500,
    alpha: float = 0.3,
    beta: float = 3,
    n_trials: int = 500,
) -> rnn_utils.DatasetRNN:
  """Generates synthetic dataset of behavior and dopamine from a reward-seeking agent."""
  np.random.seed(0)
  agent = AgentRewardSeeking(alpha=alpha, beta=beta)
  environment = two_armed_bandits.EnvironmentBanditsDrift(sigma=0.1)
  dataset = create_dataset(
      agent,
      environment,
      n_steps_per_session=n_trials,
      n_sessions=n_sessions,
      batch_size=n_sessions,
  )
  return dataset



# Copyright 2025 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Load datasets."""

import json
import os
from typing import Optional, Literal, cast
import urllib.request

from disentangled_rnns.library import pclicks
from disentangled_rnns.library import rnn_utils
from disentangled_rnns.library import two_armed_bandits
import numpy as np
import scipy.io


def find(s, ch):
  """Find index of character within string."""
  return [i for i, ltr in enumerate(s) if ltr == ch]


def get_rat_bandit_dataset(rat_i: int = 0) -> rnn_utils.DatasetRNN:
  """Downloads and packages rat two-armed bandit datasets.

  Dataset is from the following paper:
  From predictive models to cognitive models: Separable behavioral processes
  underlying reward learning in the rat. Miller, Botvinick, and Brody,
  bioRxiv, 2018

  Dataset is available from Figshare at the following link:
  https://figshare.com/articles/dataset/From_predictive_models_to_cognitive_models_Separable_behavioral_processes_underlying_reward_learning_in_the_rat/20449356

  Args:
    rat_i: Integer specifying which rat to load.

  Returns:
    A list of DatasetRNN objects. One element per rat.
    In each of these, each session will be an episode, padded with -1s
    to match length. "ys" will be the choices on each trial
    (left=0, right=1) "xs" will be the choice and reward (0 or 1) from
    the previous trial. Invalid xs and ys will be -1
  """

  # Download the file to the current folder
  url = 'https://figshare.com/ndownloader/files/40442660'
  data_path = 'tab_dataset.json'
  urllib.request.urlretrieve(url, data_path)

  # Load dataset into memory
  with open(data_path, 'r') as f:
    dataset = json.load(f)

  # Clean up after ourselves by removing the downloaded file
  os.remove(data_path)

  # "dataset" will be a list in which each element is a dict. Each of these
  # dicts holds data from a single rat.

  ratdata = dataset[rat_i]
  # "sides" is a list of characters in which each character specifies the
  # choice made on a trial. 'r' for right, 'l' for left, 'v' for a violation
  # Here, we'll code left choices as 0s, right choices as 1s, and will flag
  # violations for later removal
  sides = ratdata['sides']
  n_trials = len(sides)
  rights = find(sides, 'r')
  choices = np.zeros(n_trials)
  choices[rights] = 1
  vs = find(sides, 'v')
  viols = np.zeros(n_trials, dtype=bool)
  viols[vs] = True

  # Free will be 0 and forced will be 1
  # "trial_types" is a list of characters, each giving the type of a trial.
  # 'f' for free-choice, 'l' for instructed-left, 'r' for instructed-right
  free = find(ratdata['trial_types'], 'f')
  instructed_choice = np.ones(n_trials)
  instructed_choice[free] = 0

  # "rewards" is a list of 1s (rewarded trials) and 0s (unrewarded trials)
  rewards = np.array(ratdata['rewards'])

  # "new_sess" is a list of 1s (trials that are the first of a new session)
  # and 0s (all other trials)
  new_sess = np.array(ratdata['new_sess'])
  n_sess = np.sum(new_sess)
  sess_starts = np.nonzero(np.concatenate((new_sess, [1])))[0]
  # We will pad each session so they all have length of the longest session
  max_session_length = np.max(np.diff(sess_starts, axis=0))

  # Instantiate blank matrices for rewards and choices.
  # size (n_trials, n_sessions, 1)
  rewards_by_session = -1 * np.ones((max_session_length, n_sess, 1))
  choices_by_session = -1 * np.ones((max_session_length, n_sess, 1))
  instructed_by_session = -1 * np.ones((max_session_length, n_sess, 1))

  # Each iteration processes one session
  for sess_i in np.arange(n_sess):
    # Get the choices, rewards, viols, and instructed for just this session
    sess_start = sess_starts[sess_i]
    sess_end = sess_starts[sess_i + 1]
    viols_sess = viols[sess_start:sess_end]
    rewards_sess = rewards[sess_start:sess_end]
    choices_sess = choices[sess_start:sess_end]
    instructed_choice_sess = instructed_choice[sess_start:sess_end]

    # Remove violation trials
    rewards_sess = np.delete(rewards_sess, viols_sess)
    choices_sess = np.delete(choices_sess, viols_sess)
    instructed_choice_sess = np.delete(instructed_choice_sess, viols_sess)
    sess_length_noviols = len(choices_sess)

    # Add them to the matrices
    rewards_by_session[0:sess_length_noviols, sess_i, 0] = rewards_sess
    choices_by_session[0:sess_length_noviols, sess_i, 0] = choices_sess
    instructed_by_session[0:sess_length_noviols, sess_i, 0] = (
        instructed_choice_sess
    )

  # Define neural network inputs:
  # for each trial the choice and reward from the previous trial.
  choice_and_reward = np.concatenate(
      (choices_by_session, rewards_by_session), axis=2
  )
  xs = np.concatenate(
      (0. * np.ones((1, n_sess, 2)), choice_and_reward), axis=0
  )  # First trial's input will arbitrarily always be 0

  # Define neural network targets:
  # choices on all free-choice trial, -1 on all instructed trials
  free_choices = choices_by_session
  free_choices[instructed_by_session == 1] = -1
  # Add a dummy target at the end -- last step has input but no target
  ys = np.concatenate((free_choices, -1*np.ones((1, n_sess, 1))), axis=0)

  # Pack into a DatasetRNN object
  dataset_rat = rnn_utils.DatasetRNN(ys=ys, xs=xs, y_type='categorical')

  return dataset_rat


def get_pclicks_dataset(rat_i: int = 0) -> rnn_utils.DatasetRNN:
  """Packages up rat poisson clicks datasets.

  Dataset is from the following paper:
  Rats and humans can optimally accumulate evidence for decision-making.
  Brunton, Botvinick, and Brody. Science, 2013

  This dataset is available at the following link:
  https://github.com/Brody-Lab/brunton_dataset

  Args:
    rat_i: Integer specifying which rat to load.

  Returns:
    A list of DatasetRNN objects. One element per rat.
    In each of these, each session will be a trial. xs has two elements per
    timestep per trial, one for binned left bup counts and the other for binned
    right bup counts. ys has one element per timestep per trial, which is -1
    (masked out) on all timesteps but the final timestep, and in the final
    timestep is either 0 or 1 to indicate the choice of the rat.
  """

  # A list of URLs for datasets from individual rats
  url_path = 'https://github.com/Brody-Lab/brunton_dataset/raw/main/data/'
  url_filenames = ['chrono_B052_rawdata.mat',
                   'chrono_B053_rawdata.mat',
                   'chrono_B065_rawdata.mat',
                   'chrono_B069_rawdata.mat',
                   'chrono_B074_rawdata.mat',
                   'chrono_B083_rawdata.mat',
                   'chrono_B090_rawdata.mat',
                   'chrono_B093_rawdata.mat',
                   'chrono_B097_rawdata.mat',
                   'chrono_B102_rawdata.mat',
                   'chrono_B103_rawdata.mat',
                   'chrono_B104_rawdata.mat',
                   'chrono_B105_rawdata.mat',
                   'chrono_B106_rawdata.mat',
                   'chrono_B107_rawdata.mat',
                   'chrono_B111_rawdata.mat',
                   'chrono_B112_rawdata.mat',
                   'chrono_B113_rawdata.mat',
                   'chrono_B115_rawdata.mat',
                  ]

  # Download the file for this rat, package a datasetRNN object
  url_filename = url_filenames[rat_i]

  # Download data file
  url = url_path + url_filename
  data_path = 'pclick_ratdata.mat'
  urllib.request.urlretrieve(url, data_path)
  # Load dataset into memory
  ratdata_mat = scipy.io.loadmat(data_path)
  # Clean up after ourselves by removing the downloaded file
  os.remove(data_path)

  n_trials = ratdata_mat['total_trials'][0, 0]

  # Get the choice of the rat. 0 for left, 1 for right
  choices = np.array(ratdata_mat['rawdata']['pokedR'][0])

  # Get the stimulus duration on each trial, in 10ms bins
  stim_dur = ratdata_mat['rawdata']['T'][0]  # In seconds
  n_stim_timesteps = np.ceil(stim_dur * 100)  # In number of 10ms bins

  # Get the click counts in each bin for each trial
  binned_left_bups = np.zeros((n_trials, 100))
  binned_right_bups = np.zeros((n_trials, 100))
  for trial_i in np.arange(n_trials):
    stim_dur = n_stim_timesteps[trial_i]
    stim_start_bin = 101 - stim_dur
    # Click times, in seconds
    left_bups_trial = ratdata_mat['rawdata']['leftbups'][0, trial_i][0]
    right_bups_trial = ratdata_mat['rawdata']['rightbups'][0, trial_i][0]
    # Assemble into binned click counts
    binned_left_bups_trial, _ = np.histogram(left_bups_trial*100,
                                             bins=np.arange(stim_dur))
    binned_right_bups_trial, _ = np.histogram(right_bups_trial*100,
                                              bins=np.arange(stim_dur))

    binned_left_bups[trial_i, stim_start_bin:] = binned_left_bups_trial
    binned_right_bups[trial_i, stim_start_bin:] = binned_right_bups_trial

  # Re-arrange into inputs (xs) and targets (ys) for training RNN
  xs = np.zeros((101, n_trials, 2))
  xs[:-1, :, 0] = np.swapaxes(binned_left_bups, 1, 0)
  xs[:-1, :, 1] = np.swapaxes(binned_right_bups, 1, 0)

  ys = -1*np.ones((101, n_trials, 1))
  ys[-1,:, 0] = choices

  dataset_rat = rnn_utils.DatasetRNN(xs, ys, y_type='categorical')

  return dataset_rat


def get_q_learning_dataset(
    alpha: float = 0.3,
    beta: float = 3.,
    sigma: float = 0.1,
    n_trials: int = 500,
    n_sessions: int = 20000,
    np_rng_seed: float = 0
) -> rnn_utils.DatasetRNN:
  """Generates synthetic dataset from Q-Learning agent, using standard parameters."""
  np.random.seed(np_rng_seed)
  agent = two_armed_bandits.AgentQ(alpha=alpha, beta=beta)
  environment = two_armed_bandits.EnvironmentBanditsDrift(sigma=sigma)
  dataset = two_armed_bandits.create_dataset(
      agent,
      environment,
      n_steps_per_session=n_trials,
      n_sessions=n_sessions,
      batch_size=n_sessions,
  )
  return dataset


def get_actor_critic_dataset(
    alpha_critic: float = 0.3,
    alpha_actor_learn: float = 1.,
    alpha_actor_forget: float = 0.05,
    sigma: float = 0.1,
    n_trials: int = 500,
    n_sessions: int = 20000,
    np_rng_seed: float = 0,
) -> rnn_utils.DatasetRNN:
  """Generates synthetic dataset from Actor-Critic agent, using standard parameters."""
  np.random.seed(np_rng_seed)
  agent = two_armed_bandits.AgentLeakyActorCritic(
      alpha_critic=alpha_critic,
      alpha_actor_learn=alpha_actor_learn,
      alpha_actor_forget=alpha_actor_forget,
  )
  environment = two_armed_bandits.EnvironmentBanditsDrift(sigma=sigma)
  dataset = two_armed_bandits.create_dataset(
      agent,
      environment,
      n_steps_per_session=n_trials,
      n_sessions=n_sessions,
      batch_size=n_sessions,
  )
  return dataset


def get_bounded_accumulator_dataset(
    n_trials: int = 200000,
    stim_duration_max: int = 50,
    stim_duration_min: int = 10,
    base_click_rate: float = 10,
    click_rate_diffs: np.ndarray = np.array([-2.5, -1, 0, 1, 2.5]),
    noise_per_click: float = 0.01,
    noise_per_timestep: float = 0.0,
    click_depression: float = 1.0,
    depression_tau: float = 8.0,
    bound: float = 2.9,
    lapse: float = 0.0,
) -> rnn_utils.DatasetRNN:
  """Generates synthetic dataset from Bounded Accumulator."""
  xs, _ = pclicks.generate_clicktrains(
      n_trials=n_trials,
      stim_duration_max=stim_duration_max,
      stim_duration_min=stim_duration_min,
      base_click_rate=base_click_rate,
      click_rate_diffs=click_rate_diffs,
  )

  decisions, _ = pclicks.drift_diffusion_model(
      xs,
      noise_per_click=noise_per_click,
      noise_per_timestep=noise_per_timestep,
      click_depression=click_depression,
      depression_tau=depression_tau,
      bound=bound,
      lapse=lapse,
  )
  ys = -1 * np.ones((stim_duration_max + 1, n_trials, 1))
  ys[-1, :, 0] = decisions
  dataset = rnn_utils.DatasetRNN(xs, ys, y_type='categorical')
  return dataset


def dataset_list_to_multisubject(
    dataset_list: list[rnn_utils.DatasetRNN],
    add_subj_id: bool = True,
) -> rnn_utils.DatasetRNN:
  """Turn a list of single-subject datasets into a multisubject dataset.

  Multisubject dataset has a new first column containing an integer subject ID.
  DisRNN in multisubject mode will convert this first to a one-hot then to a
  subject embedding.

  Args:
    dataset_list: List of single-subject datasets
    add_subj_id: Whether to add a subject ID column to the xs. If True, dataset
      is suitable for multisubject mode. If False, dataset is suitable for
      single-subject mode, treating all data as if from a single subject.

  Returns:
    A single DatasetRNN containing data from all datasets in the list
  """
  xs_dataset, ys_dataset = dataset_list[0].get_all()
  x_names = dataset_list[0].x_names
  y_names = dataset_list[0].y_names
  y_type_str = dataset_list[0].y_type
  n_classes = dataset_list[0].n_classes

  # Runtime check for y_type_str before casting
  allowed_y_types = ('categorical', 'scalar', 'mixed')
  if y_type_str not in allowed_y_types:
    raise ValueError(
        f'Invalid y_type "{y_type_str}" found in dataset_list. '
        f'Expected one of {allowed_y_types}.')
  # Cast for pytype
  y_type = cast(Literal['categorical', 'scalar', 'mixed'], y_type_str)

  # If we're adding a subject ID, we'll add a feature to the xs
  if add_subj_id:
    x_dim = np.shape(xs_dataset)[2] + 1
  else:
    x_dim = np.shape(xs_dataset)[2]
  y_dim = np.shape(ys_dataset)[2]

  xs = np.zeros((0, 0, x_dim))
  ys = np.zeros((0, 0, y_dim))

  max_n_trials = 0

  # For each dataset in the list, add the subject ID column, then add it to the
  # multisubject dataset
  for dataset_i in range(len(dataset_list)):
    # Check datasets are compatible
    assert x_names == dataset_list[dataset_i].x_names, (
        f'x_names do not match across datasets. Expected {x_names}, got'
        f' {dataset_list[dataset_i].x_names}'
    )
    assert y_names == dataset_list[dataset_i].y_names, (
        f'y_names do not match across datasets. Expected {y_names}, got'
        f' {dataset_list[dataset_i].y_names}'
    )
    assert y_type == dataset_list[dataset_i].y_type, (
        f'y_type does not match across datasets. Expected {y_type}, got'
        f' {dataset_list[dataset_i].y_type}'
    )
    assert n_classes == dataset_list[dataset_i].n_classes, (
        f'n_classes does not match across datasets. Expected {n_classes}, got'
        f' {dataset_list[dataset_i].n_classes}'
    )

    xs_dataset, ys_dataset = dataset_list[dataset_i].get_all()
    n_sessions = np.shape(xs_dataset)[1]
    n_trials = np.shape(xs_dataset)[0]
    if add_subj_id:
      subj_ids = dataset_i * np.ones([n_trials, n_sessions, 1])
      xs_dataset = np.concatenate([subj_ids, xs_dataset], axis=2)

    # If this dataset has more trials than all previous datasets, add dummy
    # trials to the previous datasets so all lengths will match
    if n_trials > max_n_trials:
      # Add dummy trials to existing xs and ys
      xs = np.concatenate(
          (xs, -1 * np.ones((n_trials - max_n_trials, np.shape(xs)[1], x_dim))),
          axis=0,
      )
      ys = np.concatenate(
          (
              ys,
              -1
              * np.ones(
                  (n_trials - max_n_trials, np.shape(ys)[1], np.shape(ys)[2])
              ),
          ),
          axis=0,
      )
      max_n_trials = n_trials
    # If this dataset has fewer trials than any previous dataset, add dummy
    # trials to this dataset so all lengths will match
    elif max_n_trials > n_trials:
      xs_dataset = np.concatenate(
          (
              xs_dataset,
              -1
              * np.ones(
                  (max_n_trials - n_trials, np.shape(xs_dataset)[1], x_dim)
              ),
          ),
          axis=0,
      )
      ys_dataset = np.concatenate(
          (
              ys_dataset,
              -1
              * np.ones((
                  max_n_trials - n_trials,
                  np.shape(ys_dataset)[1],
                  np.shape(ys_dataset)[2],
              )),
          ),
          axis=0,
      )

    # Concatenate new xs and ys
    xs = np.concatenate((xs, xs_dataset), axis=1)
    ys = np.concatenate((ys, ys_dataset), axis=1)

  if add_subj_id:
    x_names = ['Subject ID'] + x_names

  dataset = rnn_utils.DatasetRNN(
      xs=xs,
      ys=ys,
      x_names=x_names,
      y_names=y_names,
      y_type=y_type,
      n_classes=n_classes,
  )

  return dataset


def get_q_learning_multisubject_dataset(
    n_trials: int = 200,
    n_sessions: int = 300,
    alphas: Optional[list[float]] = None,
    np_rng_seed: float = 0,
) -> rnn_utils.DatasetRNN:
  """Returns a multisubject dataset for the Q-learning task."""
  if alphas is None:
    alphas = [0.1, 0.2, 0.3, 0.5, 0.5, 0.6, 0.7, 0.8, 0.9]
  dataset_list = []
  for alpha in alphas:
    dataset_list.append(
        get_q_learning_dataset(
            n_trials=n_trials,
            n_sessions=n_sessions,
            alpha=alpha,
            np_rng_seed=np_rng_seed,
        )
    )
  return dataset_list_to_multisubject(dataset_list)


def get_rat_bandit_multisubject_dataset(
    n_rats: int = 20,
) -> rnn_utils.DatasetRNN:
  """Returns a multisubject dataset for the rat bandit task."""
  dataset_list = []
  for rat_i in range(n_rats):
    dataset = get_rat_bandit_dataset(rat_i=rat_i)
    dataset_list.append(dataset)
  return dataset_list_to_multisubject(dataset_list)


def get_pclick_multisubject_dataset(
    n_rats: int = 19,
) -> rnn_utils.DatasetRNN:
  """Returns a multisubject dataset for the pClick task."""
  dataset_list = []
  for rat_i in range(n_rats):
    dataset = get_pclicks_dataset(rat_i=rat_i)
    dataset_list.append(dataset)
  return dataset_list_to_multisubject(dataset_list)



# Copyright 2025 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DisRNN extension. Converts subject ID into a disentangled embedding."""

import dataclasses
from typing import Any

from disentangled_rnns.library import disrnn
import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np


@dataclasses.dataclass
class MultisubjectDisRnnConfig(disrnn.DisRnnConfig):
  """Extends DisRnnConfig with multisubject-specific parameters.

  Attributes:
   max_n_subjects: If multisubject_mode is True, this is the maximum number of
     subjects in the dataset. Otherwise must be zero
   subject_embedding_size: If multisubject_mode is True, this is the number of
     dimensions in the subject embedding. Otherwise must be zero
   subject_embedding_penalty_scale: Multiplier for KL cost on subject embedding
     bottleneck. This refers to the global subject embedding bottleneck.
  """

  max_n_subjects: int = 0
  subject_embedding_size: int = 0
  subj_penalty: float = 0.0
  update_net_subj_penalty: float = 0.0
  choice_net_subj_penalty: float = 0.0

  def __post_init__(self):
    # If the first observation is a subject ID, remove it from the list of
    # observation names before calling the base class's __post_init__
    if self.x_names is not None and self.x_names[0] == 'Subject ID':
      self.x_names = self.x_names[1:]
    super().__post_init__()
    # We expect that the first element of the input will be an integer subject
    # ID. Add this to the list of observation names to make it explicit.
    self.x_names = ['Subject ID'] + self.x_names


class MultisubjectDisRnn(disrnn.HkDisentangledRNN):
  """Disentangled RNN with multisubject mode."""

  def __init__(self, config: MultisubjectDisRnnConfig):

    # Set up subject embedding parameters
    self._max_n_subjects = config.max_n_subjects
    self._subject_embedding_size = config.subject_embedding_size
    self._subj_penalty = config.subj_penalty
    self._update_net_subj_penalty = config.update_net_subj_penalty
    self._choice_net_subj_penalty = config.choice_net_subj_penalty

    # IMPORTANT: super().__init__, will call helper method _get_haiku_parameters
    # which we have overridden below. This way of doing things is necessary for
    # Haiku to create the parameters correctly.
    super().__init__(config)
    self._get_haiku_parameters()

  def _get_haiku_parameters(self):
    """Initializes all haiku parameters for multisubject DisRNN."""
    # Build the update and choice bottlenecks, which now include the subject
    # embedding in their inputs.
    self._build_subj_emb_global_bottleneck()
    self._build_multisubject_update_bottlenecks()
    self._build_multisubject_choice_bottlenecks()
    # Build the latent bottlenecks, which are unchanged from the base class.
    super()._build_latent_bottlenecks()
    # Build the latent initializations, which are unchanged from the base class.
    self._latent_inits = hk.get_parameter(
        'latent_inits',
        (self._latent_size,),
        init=hk.initializers.RandomUniform(minval=-1, maxval=1),
    )

  def _build_subj_emb_global_bottleneck(self):
    """Initializes subject embedding bottleneck parameters."""
    # Input: subject_embedding
    # This bottleneck is applied directly to the subject_embedding.
    self._subj_emb_global_sigma, _ = (
        disrnn.get_initial_bottleneck_params(
            shape=(self._subject_embedding_size,),
            name='subj_emb_global',
        )
    )

  def _build_multisubject_update_bottlenecks(self):
    """Initializes update net bottleneck parameters, including subject embedding.
    """
    # Input: subject_embedding + observations + latents
    # Needs _obs_size from base class config to be set first.
    self._update_net_subj_sigmas, self._update_net_subj_multipliers = (
        disrnn.get_initial_bottleneck_params(
            shape=(self._subject_embedding_size, self._latent_size),
            name='update_net_subj',
        )
    )
    self._update_net_obs_sigmas, self._update_net_obs_multipliers = (
        disrnn.get_initial_bottleneck_params(
            shape=(self._obs_size, self._latent_size),
            name='update_net_obs',
        )
    )
    self._update_net_latent_sigmas, self._update_net_latent_multipliers = (
        disrnn.get_initial_bottleneck_params(
            shape=(self._latent_size, self._latent_size),
            name='update_net_latent',
        )
    )

  def _build_multisubject_choice_bottlenecks(self):
    """Initializes choice MLP bottleneck parameters, including subject embedding.
    """
    # Input: subject_embedding + latents (output from update_latents)
    self._choice_net_latent_sigmas, self._choice_net_latent_multipliers = (
        disrnn.get_initial_bottleneck_params(
            shape=(self._latent_size,),
            name='choice_net_latent',
        )
    )
    self._choice_net_subj_sigmas, self._choice_net_subj_multipliers = (
        disrnn.get_initial_bottleneck_params(
            shape=(self._subject_embedding_size,),
            name='choice_net_subj',
        )
    )

  def __call__(self, inputs: jnp.ndarray, prev_latents: jnp.ndarray):
    # Initial penalty values. Shape is (batch_size,)
    batch_size = inputs.shape[0]
    penalty = jnp.zeros(shape=(batch_size,))

    subject_id = inputs[:, 0]
    observations = inputs[:, 1:]

    # Compute the subject embeddings
    # TODO(kevinjmiller): Figure out how to add an assert here
    subject_one_hot = jax.nn.one_hot(subject_id, self._max_n_subjects)
    subject_embeddings = hk.Linear(
        self._subject_embedding_size, name='subject_embedding_weights'
    )(
        subject_one_hot,
    )  # Shape: (batch_size, subject_embedding_size)

    # Pass subject embeddings through a global bottleneck
    subject_embeddings, kl_cost = disrnn.information_bottleneck(
        inputs=subject_embeddings,
        sigmas=self._subj_emb_global_sigma,
        noiseless_mode=self._noiseless_mode,
    )
    penalty += self._subj_penalty * kl_cost

    # Set up update net inputs. These will include subject embeddings,
    # observations, and previous latents.
    # Subject embeddings for update net input
    # (batch_size, subj_emb_size, latent_size)
    subj_emb_for_update_net = jnp.tile(
        jnp.expand_dims(subject_embeddings, 2), (1, 1, self._latent_size)
    )
    subj_emb_for_update_net, kl_cost = disrnn.information_bottleneck(
        inputs=subj_emb_for_update_net,
        sigmas=self._update_net_subj_sigmas,
        multipliers=self._update_net_subj_multipliers,
        noiseless_mode=self._noiseless_mode,
    )
    penalty += self._update_net_subj_penalty * kl_cost
    # Observations for update net input (batch_size, obs_size, latent_size)
    obs_for_update_net = jnp.tile(
        jnp.expand_dims(observations, 2), (1, 1, self._latent_size)
    )
    obs_for_update_net, kl_cost = disrnn.information_bottleneck(
        inputs=obs_for_update_net,
        sigmas=self._update_net_obs_sigmas,
        multipliers=self._update_net_obs_multipliers,
        noiseless_mode=self._noiseless_mode,
    )
    penalty += self._update_net_obs_penalty * kl_cost
    # Latents for update net input
    # (batch_size, latent_size, latent_size)
    prev_latents_for_update_net = jnp.tile(
        jnp.expand_dims(prev_latents, 2), (1, 1, self._latent_size)
    )
    prev_latents_for_update_net, kl_cost = disrnn.information_bottleneck(
        inputs=prev_latents_for_update_net,
        sigmas=self._update_net_latent_sigmas,
        multipliers=self._update_net_latent_multipliers,
        noiseless_mode=self._noiseless_mode,
    )
    penalty += self._update_net_latent_penalty * kl_cost
    # Concatenate all inputs for the update networks
    # Shape: (batch_size, subj_emb_size + obs_size + latent_size, latent_size)
    update_net_inputs = jnp.concatenate(
        (
            subj_emb_for_update_net,
            obs_for_update_net,
            prev_latents_for_update_net,
        ),
        axis=1,
    )
    # Use update networks to compute new latents
    new_latents, update_net_penalty = super().update_latents(
        update_net_inputs, prev_latents
    )
    penalty += update_net_penalty

    # Predict targets
    # Subject embeddings for choice net input: (batch_size, subj_emb_size,)
    subj_emb_for_choice_net, subj_emb_kl_cost = disrnn.information_bottleneck(
        inputs=subject_embeddings,
        sigmas=self._choice_net_subj_sigmas,
        multipliers=self._choice_net_subj_multipliers,
        noiseless_mode=self._noiseless_mode,
    )
    penalty += self._choice_net_subj_penalty * subj_emb_kl_cost
    # Latents for choice net input: (batch_size, latent_size)
    latents_for_choice_net, latent_kl_cost = disrnn.information_bottleneck(
        inputs=new_latents,
        sigmas=self._choice_net_latent_sigmas,
        multipliers=self._choice_net_latent_multipliers,
        noiseless_mode=self._noiseless_mode,
    )
    penalty += self._choice_net_latent_penalty * latent_kl_cost
    # Choice net input: concatenate subject embeddings and latents
    choice_net_inputs = jnp.concatenate(
        (subj_emb_for_choice_net, latents_for_choice_net), axis=1
    )
    # Use choice network to predict targets
    predicted_targets, choice_net_penalty = super().predict_targets(
        choice_net_inputs
    )
    penalty += choice_net_penalty

    # Format the output. Shape is (batch_size, output_size + 1).
    output_shape = (batch_size, self._output_size + 1)
    output = jnp.zeros(output_shape)
    output = output.at[:, :-1].set(predicted_targets)
    output = output.at[:, -1].set(penalty)

    return output, new_latents


def get_auxiliary_metrics(
    params: hk.Params,
    open_thresh: float = 0.1,
    closed_thresh: float = 0.9,
) -> dict[str, Any]:
  """Computes auxiliary metrics for Multisubject DisRNN."""
  assert 'multisubject_dis_rnn' in params, (
      'This is the auxiliary metrics function for Multisubject DisRNN, but'
      ' params is not a MultisubjectDisRnn params. Instead it contains'
      f' {params.keys()}'
  )
  module_params = params['multisubject_dis_rnn']

  def _count_states(sigmas_array: np.ndarray,
                    open_thresh: float, closed_thresh: float):
    is_open = np.sum(sigmas_array < open_thresh)
    is_closed = np.sum(sigmas_array > closed_thresh)
    return int(is_open), int(is_closed)

  # Latent bottlenecks
  latent_s = disrnn.reparameterize_sigma(module_params['latent_sigma_params'])
  lat_open, lat_closed = _count_states(
      np.array(latent_s), open_thresh, closed_thresh)

  # Update bottlenecks (sum over subj, obs, latent inputs)
  upd_subj_s = disrnn.reparameterize_sigma(
      module_params['update_net_subj_sigma_params'])
  upd_obs_s = disrnn.reparameterize_sigma(
      module_params['update_net_obs_sigma_params'])
  upd_lat_s = disrnn.reparameterize_sigma(
      module_params['update_net_latent_sigma_params'])

  us_o, us_c = _count_states(
      np.array(upd_subj_s), open_thresh, closed_thresh)
  uo_o, uo_c = _count_states(
      np.array(upd_obs_s), open_thresh, closed_thresh)
  ul_o, ul_c = _count_states(
      np.array(upd_lat_s), open_thresh, closed_thresh)

  update_open = us_o + uo_o + ul_o
  update_closed = us_c + uo_c + ul_c

  # Choice bottlenecks (sum over subj, latent inputs)
  choice_subj_s = disrnn.reparameterize_sigma(
      module_params['choice_net_subj_sigma_params'])
  choice_lat_s = disrnn.reparameterize_sigma(
      module_params['choice_net_latent_sigma_params'])

  cs_o, cs_c = _count_states(
      np.array(choice_subj_s), open_thresh, closed_thresh)
  cl_o, cl_c = _count_states(
      np.array(choice_lat_s), open_thresh, closed_thresh)

  choice_open = cs_o + cl_o
  choice_closed = cs_c + cl_c

  bottleneck_dict = {
      'latent_bottlenecks_open': lat_open,
      'latent_bottlenecks_closed': lat_closed,
      'choice_bottlenecks_open': choice_open,
      'choice_bottlenecks_closed': choice_closed,
      'update_bottlenecks_open': update_open,
      'update_bottlenecks_closed': update_closed,
  }

  # Get total sigma (the sum of all bottleneck sigmas)
  all_sigmas_sum = (
      jnp.sum(latent_s) + jnp.sum(upd_subj_s) + jnp.sum(upd_obs_s) +
      jnp.sum(upd_lat_s) + jnp.sum(choice_subj_s) + jnp.sum(choice_lat_s)
      )
  total_sigma_val = float(all_sigmas_sum)

  return {'total_sigma': total_sigma_val, **bottleneck_dict}



# Copyright 2025 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Disentangled RNN and plotting functions for choice and neural data."""

from collections.abc import Callable
import copy
import dataclasses
from typing import Optional

from disentangled_rnns.library import disrnn
from disentangled_rnns.library import plotting
from disentangled_rnns.library import rnn_utils
import haiku as hk
import jax
import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np


@dataclasses.dataclass
class DisRnnWNeuralActivityConfig(disrnn.DisRnnConfig):
  """Specifies configuration for a Disentangled RNN with neural_activity prediction.

  Attributes:
    neural_activity_net_n_units_per_layer: Number of units per layer in the
      neural_activity network
    neural_activity_net_n_layers: Number of layers in the neural_activity net
      targets bottlenecks
  """

  neural_activity_net_n_units_per_layer: int = 8
  neural_activity_net_n_layers: int = 2
  neural_activity_net_latent_penalty: float = 1.0


class HkNeuroDisentangledRNN(disrnn.HkDisentangledRNN):
  """Disentangled RNN with an extra target for neural_activity.

  DisRNN typically have the choice prediction as the output. This class adds an
  extra output which predicts the neural_activity. Currently the neural_activity
  is expected to be a scalar floating point value.
  """

  def __init__(
      self,
      config: DisRnnWNeuralActivityConfig,
  ):
    # We are subclassing the classic DisRNN for this model. Classic disRNN is
    # used as is for the choice prediction.
    # On top of those latents we add new nets to predict the neural_activity.
    if config.obs_size != 2:
      raise ValueError('Only implemented for observation size of 2')

    super().__init__(config)
    self._neural_activity_net_n_units_per_layer = (
        config.neural_activity_net_n_units_per_layer
    )
    self._neural_activity_net_n_layers = config.neural_activity_net_n_layers
    self._neural_activity_net_latent_penalty = (
        config.neural_activity_net_latent_penalty
    )
    self._num_predictions = config.output_size + 1  # +1 for neural_activity
    self._get_haiku_parameters()

  def _build_neural_activity_bottlenecks(self):
    """Initializes parameters for the neural_activity net."""
    input_size = self._latent_size + 2
    self._neural_activity_net_sigmas, self._neural_activity_net_multipliers = (
        disrnn.get_initial_bottleneck_params(
            shape=(input_size,),
            name='neural_activity_net',
        )
    )

  def _get_haiku_parameters(self):
    """Initializes all haiku parameters for this network."""
    super()._get_haiku_parameters()
    self._build_neural_activity_bottlenecks()

  def __call__(self, observations: jnp.ndarray, prev_latents: jnp.ndarray):
    # Initial penalty values. Shape is (batch_size,)
    batch_size = observations.shape[0]
    penalty = jnp.zeros(shape=(batch_size,))

    curr_choices = observations[:, 0]
    curr_rewards = observations[:, 1]
    # Convention: First two columns are choice and reward.
    # Right now we don't have anything else in the observation but this allows
    # us to add more easily.
    observations = observations[:, 0:2]
    choice_disrnn_output, new_latents = super().__call__(
        observations, prev_latents
    )

    next_choice_prediction = choice_disrnn_output[:, :-1]
    penalty += choice_disrnn_output[:, -1]

    ####################
    # Neural Activity Net #
    ####################
    # Predict neural_activity for current time step
    # The input to the net are the latents, curr choice and curr reward.
    # The expected output of the net is the current neural_activity value.
    latents_and_curr_obs = jnp.concatenate(
        (new_latents, curr_choices[..., None], curr_rewards[..., None]), axis=1
    )
    # Apply multipliers. We want to do this whether or not we are in noiseless
    # mode, so that the model will produce similar outputs in both modes.
    neural_activity_net_inputs, neural_activity_net_kl_cost = (
        disrnn.information_bottleneck(
            inputs=latents_and_curr_obs,
            sigmas=self._neural_activity_net_sigmas,
            multipliers=self._neural_activity_net_multipliers,
            noiseless_mode=self._noiseless_mode,
        )
    )
    penalty += (
        neural_activity_net_kl_cost * self._neural_activity_net_latent_penalty
    )

    current_neural_activity_prediction, neural_activity_net_l2 = disrnn.ResMLP(
        input_size=neural_activity_net_inputs.shape[1],
        # NOTE: Neural Activity net outputs single scalars for now.
        output_size=1,
        n_units_per_layer=self._neural_activity_net_n_units_per_layer,
        n_layers=self._neural_activity_net_n_layers,
        activation_fn=self._activation,
        name='neural_activity_net',
    )(neural_activity_net_inputs)
    # Neural Activity prediction should be 1D
    current_neural_activity_prediction = jnp.squeeze(
        current_neural_activity_prediction, axis=1
    )

    if not self._noiseless_mode:
      penalty += self._l2_scale * neural_activity_net_l2

    # Note that the final output is the next choice predictions and the current
    # neural_activity prediction. The two outputs don't correspond to the same
    # trial. Neural Activity prediction is for trial n and choice prediction is
    # for trial n+1.
    output_shape = (batch_size, self._num_predictions + 1)

    output = jnp.zeros(output_shape)
    output = output.at[:, :-2].set(next_choice_prediction)
    output = output.at[:, -2].set(current_neural_activity_prediction)
    output = output.at[:, -1].set(penalty)

    return output, new_latents


def plot_bottlenecks(
    params: hk.Params,
    disrnn_w_neural_activity_config: DisRnnWNeuralActivityConfig,
    sort_latents: bool = True,
) -> plt.Figure:
  """Plot the bottleneck sigmas from an hk.DisentangledRNN."""
  params = params['hk_neuro_disentangled_rnn']
  latent_dim = params['latent_sigma_params'].shape[0]

  latent_sigmas = np.array(
      disrnn.reparameterize_sigma(params['latent_sigma_params'])
  )
  neural_activity_sigmas = np.array(
      disrnn.reparameterize_sigma(
          np.transpose(params['neural_activity_net_sigma_params'])
      )
  )

  if sort_latents:
    latent_sigma_order = np.argsort(latent_sigmas)
    neural_activity_sigmas = np.concatenate((
        neural_activity_sigmas[latent_sigma_order],
        neural_activity_sigmas[-2:],
    ))

  latent_names = np.concatenate(
      (np.arange(1, latent_dim + 1), ['prev choice', 'prev reward'])
  )

  # Plot bottlenecks from base network

  fig = plotting.plot_bottlenecks(
      {'hk_disentangled_rnn': params},
      disrnn_w_neural_activity_config,
      sort_latents,
  )

  plt.close()

  # Regenerate the figure with the neural_activity bottlenecks added to the end.
  base_axes = fig.axes
  fig, axes = plt.subplots(1, 4, figsize=(25, 5))

  im = axes[-1].imshow(
      np.swapaxes([1 - neural_activity_sigmas], 0, 1), cmap='Oranges'
  )
  im.set_clim(vmin=0, vmax=1)
  axes[-1].set_title('Neural Activity Bottlenecks')
  axes[-1].set_ylabel('Latent # (Sorted)' if sort_latents else 'Latent #')
  axes[-1].set_xticks([])  # Remove x-axis ticks as it's a 1D representation
  axes[-1].set_yticks(ticks=range(len(latent_names)), labels=latent_names)

  for i, ax in enumerate(base_axes):
    if len(ax.images) < 1:
      continue
    image = ax.images[0].get_array()
    im = axes[i].imshow(image, cmap='Oranges')
    im.set_clim(vmin=0, vmax=1)
    axes[i].set_title(ax.get_title())
    axes[i].set_ylabel(ax.get_ylabel())
    axes[i].set_xticks(
        ticks=ax.get_xticks(),
        labels=ax.get_xticklabels(),
        rotation='vertical',
    )
    axes[i].set_yticks(ticks=ax.get_yticks(), labels=ax.get_yticklabels())
    axes[i].set_ylim(ax.get_ylim())
    axes[i].set_xlim(ax.get_xlim())

  # fig.tight_layout()
  return fig


def plot_neural_activity_rules(
    params: hk.Params,
    disrnn_config: DisRnnWNeuralActivityConfig,
    axis_lim: float = 2.1,
) -> Optional[plt.Figure]:
  """Plots the neural_activity rule of a DisRNN with neural_activity prediction.

  This function visualizes how the predicted neural_activity level changes based
  on
  the model's latents, the choice made, and the reward received. It creates a
  2x2 grid of plots for different combinations of choice and reward. Within each
  subplot, it shows the neural_activity prediction as a function of the most
  influential latent variables.

  Args:
    params: The parameters of the DisRNN.
    disrnn_config: A DisRnnWNeuralActivityConfig object, expected to have
      neural_activity-related attributes.
    axis_lim: The axis limit for the plot.

  Returns:
    A matplotlib Figure object, or None if no inputs are influential.
  """

  config = copy.deepcopy(disrnn_config)
  config.noiseless_mode = True

  params_prefix = 'hk_neuro_disentangled_rnn'
  if params_prefix not in params:
    raise ValueError(f"'{params_prefix}' not found in params. ")
  params_disrnn = params[params_prefix]
  # Get neural_activity-specific attributes from config.
  try:
    neural_activity_net_n_units_per_layer = (
        config.neural_activity_net_n_units_per_layer
    )
    neural_activity_net_n_layers = config.neural_activity_net_n_layers
  except AttributeError:
    print(
        'Neural Activity net attributes not found in config. Is this a'
        ' DisRnnWNeuralActivityConfig?'
    )
    return None

  activation_fn = getattr(jax.nn, config.activation)
  latent_size = config.latent_size

  # Get sigmas and multipliers for the neural_activity network's inputs.
  neural_activity_sigmas = disrnn.reparameterize_sigma(
      params_disrnn['neural_activity_net_sigma_params']
  )
  neural_activity_multipliers = params_disrnn['neural_activity_net_multipliers']

  # Identify influential latents (sigmas < 0.5).
  influential_latents = np.where(neural_activity_sigmas[:latent_size] < 0.1)[0]
  n_influential_latents = len(influential_latents)

  if n_influential_latents == 0:
    print(
        'Neural Activity rule: No latents have a'
        ' neural_activity_net_input_sigma < 0.5. Plotting not possible.'
    )
    return None

  # Define the forward pass for the neural_activity network.
  def forward(xs):
    # Apply the same bottleneck function used during training
    neural_activity_net_inputs, _ = disrnn.information_bottleneck(
        inputs=xs,
        sigmas=neural_activity_sigmas,
        multipliers=neural_activity_multipliers,
        noiseless_mode=True,  # Keep noiseless for plotting
    )
    neural_activity_net = disrnn.ResMLP(
        input_size=neural_activity_net_inputs.shape[1],
        output_size=1,
        n_units_per_layer=neural_activity_net_n_units_per_layer,
        n_layers=neural_activity_net_n_layers,
        activation_fn=activation_fn,
        name='neural_activity_net',
    )
    prediction, _ = neural_activity_net(neural_activity_net_inputs)
    return jnp.squeeze(prediction, axis=-1)

  model = hk.transform(forward)
  apply = jax.jit(model.apply)
  neural_activity_net_params = {
      'neural_activity_net': params[params_prefix + '/neural_activity_net']
  }

  # Plotting logic.
  fig, axes = plt.subplots(
      2, 2, figsize=(10, 8), sharex=True, sharey=True, constrained_layout=True
  )
  small = 8
  medium = 12
  large = 16
  fig.suptitle('Neural Activity Prediction vs. Latents', fontsize=large)

  n_vals = 50
  latent_vals = np.linspace(-axis_lim, axis_lim, n_vals)

  # Sort influential latents by their sigma values to plot the most important.
  sorted_influential_latents = sorted(
      influential_latents, key=lambda i: neural_activity_sigmas[i]
  )
  n_latents_to_plot = min(n_influential_latents, 2)
  varying_latents_plot_indices = sorted_influential_latents[:n_latents_to_plot]

  if n_influential_latents > 2:
    print(
        'WARNING: More than two latents are influential for neural_activity.'
        ' Plotting only the two with the smallest sigmas.'
    )

  if n_latents_to_plot == 1:
    fig, ax = plt.subplots(figsize=(10, 8))
    latent_idx = varying_latents_plot_indices[0]

    for choice_val, reward_val in [(0, 0), (0, 1), (1, 0), (1, 1)]:
      base_inputs = np.zeros(latent_size + 2)
      base_inputs[latent_size] = choice_val
      base_inputs[latent_size + 1] = reward_val

      xs = np.tile(base_inputs, (n_vals, 1))
      xs[:, latent_idx] = latent_vals
      neural_activity_preds = apply(
          neural_activity_net_params, jax.random.PRNGKey(0), xs
      )

      ax.plot(
          latent_vals,
          neural_activity_preds,
          label=f'Choice={choice_val}, Reward={reward_val}',
      )

    ax.set_xlabel(f'Latent {latent_idx + 1}', fontsize=medium)
    ax.set_ylabel('Neural Activity', fontsize=medium)
    ax.set_title('Neural Activity Prediction vs. Latent', fontsize=large)
    ax.legend(fontsize=small)
    ax.grid(True, linestyle='--', alpha=0.6)
    fig.suptitle('Neural Activity Prediction vs. Latents', fontsize=large)
    return fig

  # This part is for n_latents_to_plot == 2
  fig, axes = plt.subplots(
      2, 2, figsize=(10, 8), sharex=True, sharey=True, constrained_layout=True
  )
  fig.suptitle('Neural Activity Prediction vs. Latents', fontsize=large)
  overall_min = None
  overall_max = None
  for i, (choice_val, reward_val) in enumerate(
      [(0, 0), (0, 1), (1, 0), (1, 1)]
  ):
    ax = axes.flatten()[i]
    # Add choice and reward as inputs to the neural_activity net.
    base_inputs = np.zeros(latent_size + 2)
    base_inputs[latent_size] = choice_val
    base_inputs[latent_size + 1] = reward_val

    latent_idx1, latent_idx2 = varying_latents_plot_indices
    xv, yv = np.meshgrid(latent_vals, latent_vals)
    xs = np.tile(base_inputs, (n_vals * n_vals, 1))
    xs[:, latent_idx1] = xv.flatten()
    xs[:, latent_idx2] = yv.flatten()
    neural_activity_preds = apply(
        neural_activity_net_params, jax.random.PRNGKey(0), xs
    )
    neural_activity_preds = neural_activity_preds.reshape((n_vals, n_vals))

    if i == 0:
      # Set the overall min and max to be a bit larger than the max absolute
      # prediction. We use the first plot to set the overall min and max as
      # the plots are all on the same scale.
      max_abs_pred = np.max(np.abs(neural_activity_preds))
      overall_min = -max_abs_pred * 1.1
      overall_max = max_abs_pred * 1.1

    im = ax.imshow(
        neural_activity_preds,
        cmap='coolwarm',
        origin='lower',
        aspect='auto',
        extent=[-axis_lim, axis_lim, -axis_lim, axis_lim],
        vmin=overall_min,
        vmax=overall_max,
    )
    if i == 1 or i == 3:  # Add colorbar to right plots
      cbar = fig.colorbar(im, ax=ax)
      cbar.ax.tick_params(labelsize=small)

    ax.set_xlabel(f'Latent {latent_idx1 + 1}', fontsize=medium)
    if i % 2 == 0:
      ax.set_ylabel(f'Latent {latent_idx2 + 1}', fontsize=medium)

    ax.set_title(f'Choice={choice_val}, Reward={reward_val}', fontsize=medium)
    ax.tick_params(axis='both', labelsize=small)
    ax.grid(True, linestyle='--', alpha=0.6)

  return fig


def plot_choice_rule(
    params: hk.Params,
    disrnn_config: DisRnnWNeuralActivityConfig,
    axis_lim: float = 2.1,
) -> Optional[plt.Figure]:
  """Plots the choice rule of a DisRNN with neural_activity prediction."""

  params = {
      key.replace('hk_neuro_disentangled_rnn', 'hk_disentangled_rnn'): value
      for key, value in params.items()
  }
  plotting.plot_choice_rule(
      params=params, disrnn_config=disrnn_config, axis_lim=axis_lim
  )


def plot_update_rules(
    params: hk.Params,
    disrnn_config: DisRnnWNeuralActivityConfig,
    axis_lim: float = 2.1,
) -> Optional[plt.Figure]:
  """Plots the update rules of a DisRNN with neural_activity prediction."""
  params = {
      key.replace('hk_neuro_disentangled_rnn', 'hk_disentangled_rnn'): value
      for key, value in params.items()
  }
  plotting.plot_update_rules(
      params=params, disrnn_config=disrnn_config, axis_lim=axis_lim
  )


def log_bottlenecks(
    params: hk.Params,
    open_thresh: float = 0.1,
    partially_open_thresh: float = 0.25,
    closed_thresh: float = 0.9,
) -> dict[str, int]:
  """Computes info about bottlenecks."""
  bnecks = disrnn.log_bottlenecks(
      params, open_thresh, partially_open_thresh, closed_thresh
  )

  params_disrnn = params['hk_disentangled_rnn']

  neural_activity_sigmas = np.array(
      disrnn.reparameterize_sigma(
          np.transpose(params_disrnn['neural_activity_net_sigma_params'])
      )
  )

  neural_activity_bottlenecks_open = np.sum(
      neural_activity_sigmas < open_thresh
  )
  neural_activity_bottlenecks_partial = np.sum(
      neural_activity_sigmas < partially_open_thresh
  )
  neural_activity_bottlenecks_closed = np.sum(
      neural_activity_sigmas > closed_thresh
  )

  bottleneck_dict = {
      **bnecks,
      'neural_activity_bottlenecks_open': int(neural_activity_bottlenecks_open),
      'neural_activity_bottlenecks_partial': int(
          neural_activity_bottlenecks_partial
      ),
      'neural_activity_bottlenecks_closed': int(
          neural_activity_bottlenecks_closed
      ),
  }
  return bottleneck_dict


def get_total_sigma(params):
  """Get sum of reparameterized sigmas of a DisRNN."""

  prev_sigma_total = disrnn.get_total_sigma(params)

  params_disrnn = params['hk_disentangled_rnn']

  neural_activity_bottlenecks = disrnn.reparameterize_sigma(
      params_disrnn['neural_activity_net_sigma_params']
  )

  return float(jnp.sum(neural_activity_bottlenecks) + prev_sigma_total)


def get_auxiliary_metrics(
    params: hk.Params,
    make_model_fn: Callable[[], hk.RNNCore],
    dataset_train: rnn_utils.DatasetRNN,
    dataset_eval: rnn_utils.DatasetRNN,
) -> dict[str, np.ndarray]:
  """Compute auxiliary metrics for DisRNN with Neural Activity."""
  xs, ys = dataset_train.get_all()
  network_outputs, _ = rnn_utils.eval_network(
      make_model_fn,
      params,
      xs,
  )
  y_hats = network_outputs[:, :, :-1]
  choice_normlik_train = rnn_utils.normalized_likelihood_and_mse(
      ys, y_hats, likelihood_weight=1.0
  )
  neural_activity_mse_train = rnn_utils.normalized_likelihood_and_mse(
      ys, y_hats, likelihood_weight=0.0
  )

  xs, ys = dataset_eval.get_all()
  network_outputs, _ = rnn_utils.eval_network(
      make_model_fn,
      params,
      xs,
  )
  y_hats = network_outputs[:, :, :-1]
  choice_normlik_eval = rnn_utils.normalized_likelihood_and_mse(
      ys, y_hats, likelihood_weight=1.0
  )
  neural_activity_mse_eval = rnn_utils.normalized_likelihood_and_mse(
      ys, y_hats, likelihood_weight=0.0
  )

  bottleneck_metrics = log_bottlenecks(params)
  total_sigma_val = get_total_sigma(params)
  total_sigma_val = np.array(total_sigma_val)

  neural_activity_specific_metrics = {
      'choice_normlik_train': np.array(choice_normlik_train),
      'choice_normlik_eval': np.array(choice_normlik_eval),
      'neural_activity_mse_train': np.array(neural_activity_mse_train),
      'neural_activity_mse_eval': np.array(neural_activity_mse_eval),
  }
  return {
      **neural_activity_specific_metrics,
      **bottleneck_metrics,
      'total_sigma': total_sigma_val,
  }



# Copyright 2025 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example DisRNN workflow: Define a dataset, train network, inspect the fit.
"""

import copy

from absl import app
from absl import flags
from disentangled_rnns.library import disrnn
from disentangled_rnns.library import plotting
from disentangled_rnns.library import rnn_utils
from disentangled_rnns.library import two_armed_bandits
import optax


FLAGS = flags.FLAGS

flags.DEFINE_integer(
    "n_steps_per_session", 200, "Number of steps per session in the dataset."
)
flags.DEFINE_integer("n_sessions", 300, "Number of sessions in the dataset.")
flags.DEFINE_float("learning_rate", 1e-3, "Optimizer learning rate.")
flags.DEFINE_integer("n_warmup_steps", 1000, "Number of training warmup steps.")
flags.DEFINE_integer(
    "n_training_steps", 3000, "Number of main training steps."
)


def main(_) -> None:

  ##############################
  # Define dataset and network #
  ##############################

  # Synthetic dataset from q-learning agent. Other options in get_datasets.py
  agent = two_armed_bandits.AgentQ(alpha=0.3, beta=3)
  environment = two_armed_bandits.EnvironmentBanditsDrift(sigma=0.1)
  dataset = two_armed_bandits.create_dataset(
      agent,
      environment,
      n_steps_per_session=FLAGS.n_steps_per_session,
      n_sessions=FLAGS.n_sessions,
      batch_size=FLAGS.n_sessions,
  )

  # Second synthetic dataset for evaluation
  dataset_eval = two_armed_bandits.create_dataset(
      agent,
      environment,
      n_steps_per_session=FLAGS.n_steps_per_session,
      n_sessions=FLAGS.n_sessions,
      batch_size=FLAGS.n_sessions,
  )

  # Define the disRNN architecture
  disrnn_config = disrnn.DisRnnConfig(
      # Dataset related
      obs_size=2,
      output_size=2,
      x_names=dataset.x_names,
      y_names=dataset.y_names,
      # Network architecture
      latent_size=5,
      update_net_n_units_per_layer=8,
      update_net_n_layers=4,
      choice_net_n_units_per_layer=4,
      choice_net_n_layers=2,
      activation="leaky_relu",
      # Penalties
      noiseless_mode=False,
      latent_penalty=1e-2,
      update_net_obs_penalty=1e-3,
      update_net_latent_penalty=1e-3,
      choice_net_latent_penalty=1e-3,
      l2_scale=1e-5,
  )
  # Define a config for warmup training with no noise and no penalties
  disrnn_config_warmup = copy.deepcopy(disrnn_config)
  disrnn_config_warmup.latent_penalty = 0
  disrnn_config_warmup.choice_net_latent_penalty = 0
  disrnn_config_warmup.update_net_obs_penalty = 0
  disrnn_config_warmup.update_net_latent_penalty = 0
  disrnn_config_warmup.l2_scale = 0
  disrnn_config_warmup.noiseless_mode = True

  # Define network builder functions
  make_disrnn = lambda: disrnn.HkDisentangledRNN(disrnn_config)
  make_disrnn_warmup = lambda: disrnn.HkDisentangledRNN(disrnn_config_warmup)

  # Define an optimizer
  opt = optax.adam(learning_rate=FLAGS.learning_rate)

  #################################
  # Optimizing network parameters #
  #################################

  # Warmup training with no noise and no penalties
  params, _, _ = rnn_utils.train_network(
      make_disrnn_warmup,
      training_dataset=dataset,
      validation_dataset=dataset_eval,
      loss="penalized_categorical",
      params=None,
      opt_state=None,
      opt=opt,
      n_steps=FLAGS.n_warmup_steps,
      do_plot=True,
  )

  # Additional training using information penalty
  params, _, _ = rnn_utils.train_network(
      make_disrnn,
      training_dataset=dataset,
      validation_dataset=dataset_eval,
      loss="penalized_categorical",
      params=params,
      opt_state=None,
      opt=opt,
      n_steps=FLAGS.n_training_steps,
      do_plot=True,
  )

  ###########################
  # Inspecting a fit disRNN #
  ###########################
  # Plot bottleneck structure and update rules
  plotting.plot_bottlenecks(params, disrnn_config)
  plotting.plot_update_rules(params, disrnn_config)

  ##############################
  # Eval disRNN on unseen data #
  ##############################
  # Use the wamrup disrnn, so that there will be no noise
  xs, _ = next(dataset_eval)
  # pylint: disable-next=unused-variable
  _, network_states = rnn_utils.eval_network(make_disrnn_warmup, params, xs)


if __name__ == "__main__":
  app.run(main)



# Copyright 2025 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Code for generating synthetic data from the poisson clicks task.
"""
import numpy as np


def generate_clicktrains(n_trials=10000,
                         base_click_rate=20,
                         click_rate_diffs=(-38, -34, -27, -12, 12, 27, 34, 38),
                         stim_duration_max=100,
                         stim_duration_min=20):
  """Generates random clicktrain stimuli for Poisson clicks task.

  Args:
    n_trials:
    base_click_rate: Average number of clicks on each side
    click_rate_diffs: Difference in click rates on different trial types
    stim_duration_max: Number of 10ms time bins. Each trial's stim duration will
      be sampled uniformly between this and stim_durtion_min
    stim_duration_min: Number of 10ms time bins

  Returns:

  """

  # Leave a blank timestep at the end for the response
  xs = np.zeros((stim_duration_max + 1, n_trials, 2))
  stim_durations = np.zeros(n_trials)

  for trial_i in range(n_trials):
    stim_duration = np.random.randint(stim_duration_min, stim_duration_max)

    diff_i = np.random.randint(0, len(click_rate_diffs))
    click_rate_diff = click_rate_diffs[diff_i]
    rate_left = (base_click_rate - click_rate_diff / 2) / stim_duration_max
    rate_right = (base_click_rate + click_rate_diff / 2) / stim_duration_max

    left_clicks = np.random.binomial(1, rate_left, (stim_duration, 1))
    right_clicks = np.random.binomial(1, rate_right, (stim_duration, 1))

    xs_trial = [left_clicks, right_clicks]
    xs_trial = np.swapaxes(xs_trial, 0, 2)

    xs[-(stim_duration + 1) : -1, trial_i, :] = xs_trial

    stim_durations[trial_i] = stim_duration

  assert np.all(xs[-1, :, :] == 0)  # Check that the last timestep has no clicks

  return xs, stim_durations


def drift_diffusion_model(
    xs,
    noise_per_click=0.5,
    noise_per_timestep=0,
    click_depression=0.3,
    depression_tau=8,
    bound=5,
    lapse=0,
):
  """Runs DDM on clicktrains to produce choices.

  Args:
    xs: Click trains, as output by generate_clicktrains
    noise_per_click: Sensory noise per click
    noise_per_timestep: Accumulator noise per timestep
    click_depression: Amount of sensory depression caused by a single click
    depression_tau: Timeconstant for recovery from click depression
    bound: Sticky decision bound, in units of clicks
    lapse: Lapse rate

  Returns:

  """
  (n_timesteps, n_trials, _) = np.shape(xs)

  click_any = np.logical_or(xs[:, :, 0], xs[:, :, 1]).astype(float)
  depression_variable = np.nan * np.ones((n_timesteps, n_trials))

  depression_variable[0, :] = 1.0
  for timestep_i in range(n_timesteps - 1):
    # First, decay the depression variable towards one
    depression_variable[timestep_i + 1, :] = (
        depression_variable[timestep_i, :]
        + (1 - depression_variable[timestep_i, :]) / depression_tau
    )
    # Next multiply it by phi if there has been a click
    depression_variable[timestep_i + 1, :] += (
        (click_depression - 1)
        * depression_variable[timestep_i, :]
        * click_any[timestep_i, :]
    )

  diffusion_noise = np.random.normal(
      loc=0, scale=noise_per_timestep, size=(n_timesteps, n_trials)
  )

  noisy_clicks_left = np.random.normal(
      loc=xs[:, :, 0],
      scale=noise_per_click * xs[:, :, 0],
      size=(n_timesteps, n_trials),
  )
  noisy_clicks_right = np.random.normal(
      loc=xs[:, :, 1],
      scale=noise_per_click * xs[:, :, 1],
      size=(n_timesteps, n_trials),
  )

  click_evidence = np.multiply(
      noisy_clicks_left - noisy_clicks_right, depression_variable
  )
  decision_variable = np.cumsum(click_evidence, axis=0) + np.cumsum(
      diffusion_noise, axis=0
  )

  crossed_bound = np.logical_or(
      decision_variable > bound, decision_variable < -1 * bound
  )
  first_bound_crossing = np.argmax(crossed_bound, axis=0)

  decisions = np.zeros(n_trials)
  for trial_i in range(n_trials):
    if first_bound_crossing[trial_i] > 0:
      decision_variable[first_bound_crossing[trial_i] :, trial_i] = (
          bound
          * np.sign(decision_variable[first_bound_crossing[trial_i], trial_i])
      )

    decision = (np.sign(decision_variable[-1, trial_i]) + 1) / 2
    decision = (decision + 1) / 2  # Convert from +-1 to 0 or 1

    if np.random.random() < lapse:
      decision = 1 - decision
    decisions[trial_i] = int(decision)

  return decisions, (decision_variable, depression_variable)
